"""Convert a Data Package to an Avro schema."""

import json
import sys
from typing import Dict, List
from datapackage import Package

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | bool | int | None


class DataPackageToAvroConverter:
    """Class to convert Data Package to Avro schema."""

    def convert_datapackage_to_avro(self, datapackage_path, avro_schema_path):
        """Convert a Data Package to an Avro schema."""
        package = Package(datapackage_path)
        resources = package.resources

        avro_schemas = []

        for resource in resources:
            table_name = resource.descriptor['name']
            fields = resource.descriptor['schema']['fields']

            avro_fields = []
            for field in fields:
                avro_field = {
                    "name": field["name"],
                    "type": self.convert_datapackage_type_to_avro_type(field["type"])
                }
                avro_fields.append(avro_field)

            avro_schema = {
                "type": "record",
                "name": table_name,
                "fields": avro_fields
            }
            avro_schemas.append(avro_schema)

        # If there's only one schema, write it directly
        if len(avro_schemas) == 1:
            avro_schema = avro_schemas[0]
        else:
            # If there are multiple schemas, create a union
            avro_schema = avro_schemas

        with open(avro_schema_path, "w", encoding="utf-8") as f:
            json.dump(avro_schema, f, indent=2)

    def convert_datapackage_type_to_avro_type(self, datapackage_type):
        """Convert a Data Package type to an Avro type."""
        if datapackage_type == "string":
            return "string"
        elif datapackage_type == "number":
            return "double"
        elif datapackage_type == "integer":
            return "int"
        elif datapackage_type == "boolean":
            return "boolean"
        elif datapackage_type == "array":
            return {"type": "array", "items": "string"}
        elif datapackage_type == "object":
            return {"type": "map", "values": "string"}
        else:
            print(f"WARNING: Unsupported data package type: {datapackage_type}")
            return "string"


def convert_datapackage_to_avro(datapackage_path, avro_schema_path):
    """Convert a Data Package to an Avro schema."""
    converter = DataPackageToAvroConverter()
    converter.convert_datapackage_to_avro(datapackage_path, avro_schema_path)


# Example usage:
# convert_datapackage_to_avro("datapackage.json", "schema.avsc")



""" Convert Avro schema to TypeScript classes """

import json
import os
from typing import Any, Dict, List, Set, Union

from avrotize.common import pascal

INDENT = ' ' * 4

def is_javascript_reserved_word(word: str) -> bool:
    """ Check if word is a TypeScript reserved word """
    reserved_words = [
        'break', 'case', 'catch', 'class', 'const', 'continue', 'debugger',
        'default', 'delete', 'do', 'else', 'export', 'extends', 'finally',
        'for', 'function', 'if', 'import', 'in', 'instanceof', 'new', 'return',
        'super', 'switch', 'this', 'throw', 'try', 'typeof', 'var', 'void',
        'while', 'with', 'yield'
    ]
    return word in reserved_words

def is_javascript_primitive(word: str) -> bool:
    """ Check if word is a TypeScript primitive """
    primitives = ['null', 'boolean', 'number', 'string', 'Date']
    return word in primitives

class AvroToJavaScript:
    """ Convert Avro schema to TypeScript classes """
    
    def __init__(self, base_package: str = '', avro_annotation=False) -> None:
        self.base_package = base_package
        self.avro_annotation = avro_annotation
        self.output_dir = os.getcwd()

    def map_primitive_to_javascript(self, avro_type: str) -> str:
        """ Map Avro primitive type to TypeScript type """
        mapping = {
            'null': 'null',
            'boolean': 'boolean',
            'int': 'number',
            'long': 'number',
            'float': 'number',
            'double': 'number',
            'bytes': 'string',
            'string': 'string',
        }
        return mapping.get(avro_type, avro_type)

    def convert_logical_type_to_javascript(self, avro_type: Dict) -> str:
        """ Convert Avro logical type to TypeScript type """
        if 'logicalType' in avro_type:
            if avro_type['logicalType'] in ['decimal', 'uuid']:
                return 'string'
            if avro_type['logicalType'] in ['date', 'time-millis', 'time-micros', 'timestamp-millis', 'timestamp-micros']:
                return 'Date'
            if avro_type['logicalType'] == 'duration':
                return 'string'
        return 'any'
    
    def is_javascript_primitive(self, avro_type: str) -> bool:
        """ Check if Avro type is a TypeScript primitive """
        return avro_type in ['null', 'boolean', 'number', 'string', 'Date']

    def convert_avro_type_to_javascript(self, avro_type: Union[str, Dict, List], parent_namespace: str, import_types: set) -> str | List[Any]:
        """ Convert Avro type to TypeScript type """
        if isinstance(avro_type, str):
            mapped_type = self.map_primitive_to_javascript(avro_type)
            if mapped_type == avro_type and not self.is_javascript_primitive(mapped_type):
                import_types.add(mapped_type)
                return pascal(mapped_type.split('.')[-1])
            return mapped_type
        elif isinstance(avro_type, list):
            return [self.convert_avro_type_to_javascript(t, parent_namespace, import_types) for t in avro_type]
        elif isinstance(avro_type, dict):
            if 'type' in avro_type and avro_type['type'] == 'record':
                class_ref = self.generate_class(avro_type, parent_namespace)
                import_types.add(class_ref)
                return class_ref.split('.')[-1]
            elif 'type' in avro_type and avro_type['type'] == 'enum':
                enum_ref = self.generate_enum(avro_type, parent_namespace)
                import_types.add(enum_ref)
                return enum_ref.split('.')[-1]
            elif 'type' in avro_type and avro_type['type'] == 'array':
                return [self.convert_avro_type_to_javascript(avro_type["items"], parent_namespace, import_types)]
            if 'type' in avro_type and avro_type['type'] == 'map':
                return [self.convert_avro_type_to_javascript(avro_type["values"], parent_namespace, import_types)]
            if 'logicalType' in avro_type:
                return self.convert_logical_type_to_javascript(avro_type)
            return self.convert_avro_type_to_javascript(avro_type['type'], parent_namespace, import_types)
        return 'any'

    def generate_class(self, avro_schema: Dict, parent_namespace: str) -> str:
        """ Generate TypeScript class from Avro record """
        
        def add_check(arr_check, ft):
            if isinstance(ft, list):
                for ft2 in ft:
                    add_check(arr_check, ft2)                
            elif ft == 'null':
                arr_check.append('v !== null')
            elif is_javascript_primitive(ft):
                arr_check.append(f'typeof v !== "{ft}"')
            else:
                arr_check.append(f'!(v instanceof {ft})')
        
        import_types: Set[str] = set()
        class_name = pascal(avro_schema['name'])
        namespace = avro_schema.get('namespace', '')
        if not namespace and parent_namespace:
            namespace = parent_namespace
        if self.base_package:
            namespace = f'{self.base_package}.{namespace}'
        fields = avro_schema.get('fields', [])
        doc = avro_schema.get('doc', '')
        
        constructor_body = ''
        class_body = ''
        if self.avro_annotation:
            avro_schema_json = json.dumps(avro_schema)
            avro_schema_json = avro_schema_json.replace('"', '§')           
            avro_schema_json = f"\"+\n{' '*8}\"".join([avro_schema_json[i:i+80] for i in range(0, len(avro_schema_json), 80)])
            avro_schema_json = avro_schema_json.replace('§', '\\"')
            class_body += f'{class_name}.AvroType = avro.parse("{avro_schema_json}");\n'
        for field in fields:
            field_name = field['name']
            field_doc = field.get('doc', '')
            field_avro_type = field['type']
            if is_javascript_reserved_word(field_name):
                field_name += '_'
            field_type = self.convert_avro_type_to_javascript(field['type'], namespace, import_types)
            if field_doc:
                constructor_body += f'{INDENT}/** {field_doc} */\n'
            constructor_body += f'{INDENT}_{field_name} = null;\n'
            class_body += f'Object.defineProperty({class_name}.prototype, "{field_name}", {{\n'
            class_body += f'{INDENT}get: function() {{'+'\n'
            class_body += f'{INDENT}{INDENT}return this._{field_name};\n'
            class_body += f'{INDENT}}},\n'
            class_body += f'{INDENT}set: function(value) {{'+'\n'
            type_check = []
            if field_avro_type == 'array':
                arr = '!Array.isArray(value) || value.some(v => '
                arr_check: List[str] = []
                for ft in field_type if isinstance(field_type, list) else [field_type]:
                    add_check(arr_check, ft)
                arr += ' || '.join(arr_check) + ')'
                type_check.append(arr)
            elif field_avro_type == 'map':
                map = '!Object.entries(value).every(([k, v]) => typeof k === "string" && '
                map_check: List[str] = []
                for ft in field_type if isinstance(field_type, list) else [field_type]:
                    add_check(map_check, ft)
                map += ' && '.join(map_check) + ')'
                type_check.append(map)
            else:
                for ft in field_type if isinstance(field_type, list) else [field_type]:
                    add_check(type_check, ft)
            class_body += f'{INDENT}{INDENT}if ( {" && ".join(type_check)} ) throw new Error(`Invalid type for {field_name}. Expected {field_type}, got ${{value}}`);\n'
            class_body += f'{INDENT}{INDENT}this._{field_name} = value;\n'
            class_body += f'{INDENT}}}'+'\n'
            class_body += '});\n\n'
        
        imports = ''
        if self.avro_annotation:
            imports += "var avro = require('avro-js');\n"
        for import_type in import_types:
            import_type_package = import_type.rsplit('.',1)[0]
            import_type_type = pascal(import_type.split('.')[-1])
            import_type_package = import_type_package.replace('.', '/')
            namespace_path = namespace.replace('.', '/')
            
            if import_type_package:# get the relative path from the namespace to the import_type_package
                import_type_package = os.path.relpath(import_type_package, namespace_path).replace(os.sep, '/')
                if not import_type_package.startswith('.'):
                    import_type_package = f'./{import_type_package}'
                imports += f"var {import_type_type} = require('{import_type_package}/{import_type_type}').{import_type_type};\n"
            else:
                imports += f"var {import_type_type} = require('{import_type_type}'){import_type_type};\n"
        
        class_definition = imports + '\n'
        if doc:
            class_definition += f'/** {doc} */\n'        
        class_definition += f"function {class_name}() {{\n{constructor_body}}}\n\n{class_body}\nmodule.exports = {class_name};\n"
        self.write_to_file(namespace, class_name, class_definition)
        return f'{namespace}.{class_name}'

    def generate_enum(self, avro_schema: Dict, parent_namespace: str) -> str:
        """ Generate TypeScript enum from Avro enum """
        enum_name = pascal(avro_schema['name'])
        namespace = avro_schema.get('namespace', '')
        if not namespace and parent_namespace:
            namespace = parent_namespace
        if self.base_package:
            namespace = f'{self.base_package}.{namespace}'
        symbols = avro_schema.get('symbols', [])
        
        enum_body = ''
        for symbol in symbols:
            if is_javascript_reserved_word(symbol):
                symbol += '_'
            enum_body += f'{INDENT}{symbol}: "{symbol}",\n'
            
        enum_definition = ''
        if 'doc' in avro_schema:
            enum_definition += f"/** {avro_schema['doc']} */\n"
        enum_definition += f"const {enum_name} = Object.freeze({{\n{enum_body}}});\n\n"
        enum_definition += f"module.exports = {enum_name};\n"
        self.write_to_file(namespace, enum_name, enum_definition)
        return f'{namespace}.{enum_name}'

    def write_to_file(self, namespace: str, name: str, content: str):
        """ Write TypeScript class to file """
        directory_path = os.path.join(self.output_dir, namespace.replace('.', os.sep))
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)
        
        file_path = os.path.join(directory_path, f"{name}.js")
        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(content)

    def convert_schema(self, schema: List|Dict, output_dir: str):
        """ Convert Avro schema to TypeScript classes """
        self.output_dir = output_dir
        if isinstance(schema, dict):
            schema = [schema]

        for avro_schema in schema:
            if avro_schema['type'] == 'record':
                self.generate_class(avro_schema, self.base_package)
            elif avro_schema['type'] == 'enum':
                self.generate_enum(avro_schema, self.base_package)
                
    def convert(self, avro_schema_path: str, output_dir: str):
        """ Convert Avro schema to TypeScript classes """
        with open(avro_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        self.convert_schema(schema, output_dir)

def convert_avro_to_javascript(avro_schema_path, js_dir_path, package_name='', avro_annotation=False):
    """ Convert Avro schema to TypeScript classes """
    
    if not package_name:
        package_name = os.path.splitext(os.path.basename(avro_schema_path))[0].replace('-', '_')
    
    converter = AvroToJavaScript(package_name, avro_annotation=avro_annotation)
    converter.convert(avro_schema_path, js_dir_path)
    
def convert_avro_schema_to_javascript(avro_schema, js_dir_path, package_name='', avro_annotation=False):
    """ Convert Avro schema to TypeScript classes """
    converter = AvroToJavaScript(package_name, avro_annotation=avro_annotation)
    converter.convert_schema(avro_schema, js_dir_path)



""" JSON Structure to JSON Schema converter. """

# pylint: disable=too-many-lines, line-too-long, too-many-branches, too-many-statements, too-many-locals, too-many-nested-blocks, too-many-arguments, too-many-instance-attributes, too-many-public-methods, too-many-boolean-expressions

import json
import os
from typing import Any, Dict, List, Union, Optional


class StructureToJsonConverter:
    """
    Converts JSON Structure documents to JSON Schema format.
    
    JSON Structure is more constrained than JSON Schema, making this conversion
    straightforward with well-defined mappings.
    """

    def __init__(self) -> None:
        """Initialize the converter."""
        self.definitions: Dict[str, Any] = {}
        self.base_uri = ""
        self.structure_document: Optional[Dict[str, Any]] = None

    def convert_type(self, structure_type: str) -> Dict[str, Any]:
        """
        Convert a JSON Structure type to JSON Schema type.
        
        Args:
            structure_type (str): The JSON Structure type
            
        Returns:
            Dict[str, Any]: JSON Schema type definition
        """
        # Basic type mappings from JSON Structure to JSON Schema
        # Note: JSON Structure types have specific serialization rules:
        # - int64, uint64, decimal are serialized as strings in JSON due to precision/range limits
        # - binary32/binary64 are IEEE 754 binary formats
        # - timestamp is Unix epoch time as number
        type_mappings = {
            'null': {'type': 'null'},
            'string': {'type': 'string'},
            'boolean': {'type': 'boolean'},
            'bytes': {'type': 'string', 'format': 'byte'},
            'int8': {'type': 'integer', 'minimum': -128, 'maximum': 127},
            'int16': {'type': 'integer', 'minimum': -32768, 'maximum': 32767},
            'int32': {'type': 'integer', 'minimum': -2147483648, 'maximum': 2147483647},
            'int64': {'type': 'string', 'pattern': '^-?[0-9]+$'},  # Serialized as string
            'uint8': {'type': 'integer', 'minimum': 0, 'maximum': 255},
            'uint16': {'type': 'integer', 'minimum': 0, 'maximum': 65535},            'uint32': {'type': 'integer', 'minimum': 0, 'maximum': 4294967295},
            'uint64': {'type': 'string', 'pattern': '^[0-9]+$'},  # Serialized as string
            'int128': {'type': 'string', 'pattern': '^-?[0-9]+$'},  # Serialized as string
            'uint128': {'type': 'string', 'pattern': '^[0-9]+$'},  # Serialized as string
            'float8': {'type': 'number'},  # 8-bit float
            'float': {'type': 'number', 'format': 'float'},
            'double': {'type': 'number', 'format': 'double'},
            'float32': {'type': 'number', 'format': 'float'},  # 32-bit float
            'float64': {'type': 'number', 'format': 'double'},  # 64-bit float
            'binary32': {'type': 'number', 'format': 'float'},  # IEEE 754 binary32 (alias)
            'binary64': {'type': 'number', 'format': 'double'}, # IEEE 754 binary64 (alias)
            'decimal': {'type': 'string', 'pattern': '^-?[0-9]+(\\.[0-9]+)?$'},  # Serialized as string
            'binary': {'type': 'string', 'contentEncoding': 'base64'},  # Binary data
            'date': {'type': 'string', 'format': 'date'},
            'time': {'type': 'string', 'format': 'time'},
            'datetime': {'type': 'string', 'format': 'date-time'},
            'timestamp': {'type': 'number'},  # Unix epoch time
            'duration': {'type': 'string', 'format': 'duration'},
            'uuid': {'type': 'string', 'format': 'uuid'},
            'uri': {'type': 'string', 'format': 'uri'},
            'jsonpointer': {'type': 'string', 'format': 'json-pointer'},
        }
        
        return type_mappings.get(structure_type, {'type': 'string'})

    def convert_structure_schema(self, structure: Dict[str, Any]) -> Dict[str, Any]:
        """
        Convert a JSON Structure schema to JSON Schema.
        
        Args:
            structure (Dict[str, Any]): The JSON Structure schema
            
        Returns:
            Dict[str, Any]: The converted JSON Schema
        """
        # Store the structure document for reference resolution
        self.structure_document = structure
        
        schema: Dict[str, Any] = {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
        }
          # Preserve $id if present
        if '$id' in structure:
            schema['$id'] = structure['$id']
            
        # Preserve $uses as extension if present
        if '$uses' in structure:
            schema['x-uses'] = structure['$uses']
            
        # Convert title and description
        if 'name' in structure:
            schema['title'] = structure['name']
        if 'description' in structure:
            schema['description'] = structure['description']
            
        # Convert the main type
        if 'type' in structure:
            schema_type = self._convert_type_definition(structure)
            schema.update(schema_type)
          # Add definitions if we collected any
        if self.definitions:
            schema['$defs'] = self.definitions
            
        return schema

    def _convert_type_definition(self, type_def: Dict[str, Any]) -> Dict[str, Any]:
        """
        Convert a JSON Structure type definition to JSON Schema.
        
        Args:
            type_def (Dict[str, Any]): The type definition
            
        Returns:
            Dict[str, Any]: JSON Schema type definition
        """
        if not isinstance(type_def, dict) or 'type' not in type_def:
            return {'type': 'string'}
            
        structure_type = type_def['type']
        
        # Handle type unions (arrays of types)
        if isinstance(structure_type, list):
            return self._convert_type_union(structure_type, type_def)
        
        if structure_type == 'object':
            return self._convert_object_type(type_def)
        elif structure_type == 'array':
            return self._convert_array_type(type_def)
        elif structure_type == 'set':
            # Sets in JSON Structure become arrays with uniqueItems in JSON Schema
            return self._convert_set_type(type_def)
        elif structure_type == 'map':
            return self._convert_map_type(type_def)
        elif structure_type == 'choice':
            return self._convert_choice_type(type_def)
        else:
            # Primitive type
            result = self.convert_type(structure_type)
            
            # Add validation constraints
            self._add_validation_constraints(result, type_def)
            
            return result

    def _convert_object_type(self, type_def: Dict[str, Any]) -> Dict[str, Any]:
        """Convert a JSON Structure object type to JSON Schema."""
        result = {'type': 'object'}
        
        if 'properties' in type_def:
            result['properties'] = {}
            for prop_name, prop_def in type_def['properties'].items():
                result['properties'][prop_name] = self._convert_type_definition(prop_def)
        
        if 'required' in type_def:
            result['required'] = type_def['required']
            
        if 'additionalProperties' in type_def:
            if isinstance(type_def['additionalProperties'], bool):
                result['additionalProperties'] = type_def['additionalProperties']
            else:
                result['additionalProperties'] = self._convert_type_definition(type_def['additionalProperties'])
        
        return result

    def _convert_array_type(self, type_def: Dict[str, Any]) -> Dict[str, Any]:
        """Convert a JSON Structure array type to JSON Schema."""
        result = {'type': 'array'}
        
        if 'items' in type_def:
            result['items'] = self._convert_type_definition(type_def['items'])
            
        # Add validation constraints
        if 'minItems' in type_def:
            result['minItems'] = type_def['minItems']
        if 'maxItems' in type_def:
            result['maxItems'] = type_def['maxItems']
            
        return result

    def _convert_set_type(self, type_def: Dict[str, Any]) -> Dict[str, Any]:
        """Convert a JSON Structure set type to JSON Schema array with uniqueItems."""
        result = {'type': 'array', 'uniqueItems': True}
        
        if 'items' in type_def:
            result['items'] = self._convert_type_definition(type_def['items'])
            
        # Add validation constraints
        if 'minItems' in type_def:
            result['minItems'] = type_def['minItems']
        if 'maxItems' in type_def:
            result['maxItems'] = type_def['maxItems']
            
        return result

    def _convert_map_type(self, type_def: Dict[str, Any]) -> Dict[str, Any]:
        """Convert a JSON Structure map type to JSON Schema object with additionalProperties."""
        result = {'type': 'object'}
        
        if 'values' in type_def:
            result['additionalProperties'] = self._convert_type_definition(type_def['values'])
        else:
            result['additionalProperties'] = True
            
        return result

    def _convert_choice_type(self, type_def: Dict[str, Any]) -> Dict[str, Any]:
        """Convert a JSON Structure choice type to JSON Schema oneOf."""
        result = {}
        
        if 'choices' in type_def:
            choices = type_def['choices']
            if isinstance(choices, dict):
                # Tagged union: choices is a dict mapping names to types
                result['oneOf'] = []
                for choice_name, choice_def in choices.items():
                    converted_choice = None
                    resolved_choice = None
                    
                    if isinstance(choice_def, dict):
                        # Handle $ref resolution
                        if '$ref' in choice_def:
                            ref_type = self._resolve_reference(choice_def['$ref'])
                            if ref_type:
                                resolved_choice = self._resolve_type_with_inheritance(ref_type)
                                converted_choice = self._convert_type_definition(resolved_choice)
                            else:
                                # Fallback if reference can't be resolved
                                converted_choice = {'type': 'string'}
                        else:
                            resolved_choice = choice_def
                            converted_choice = self._convert_type_definition(choice_def)
                    elif isinstance(choice_def, str):
                        # Simple type name
                        converted_choice = self.convert_type(choice_def)
                    else:
                        # Fallback to string type
                        converted_choice = {'type': 'string'}
                    
                    # For tagged unions, wrap in an object with the choice name
                    if not type_def.get('selector'):  # Tagged union
                        tagged_choice = {
                            'type': 'object',
                            'properties': {
                                choice_name: converted_choice
                            },
                            'required': [choice_name],
                            'additionalProperties': False
                        }
                        result['oneOf'].append(tagged_choice)
                    else:                        # Inline union with selector - resolve the base type and merge with choices
                        if '$extends' in type_def:
                            # Resolve the base type that provides common properties
                            base_type = self._resolve_reference(type_def['$extends'])
                            if base_type:
                                resolved_base = self._resolve_type_with_inheritance(base_type)
                                # Merge base properties with the choice-specific properties
                                if resolved_choice and 'properties' in resolved_base:
                                    choice_properties = {}
                                    base_props = resolved_base.get('properties', {})
                                    choice_props = resolved_choice.get('properties', {})
                                    
                                    # Merge properties from base and choice
                                    if isinstance(base_props, dict):
                                        choice_properties.update(base_props)
                                    if isinstance(choice_props, dict):
                                        choice_properties.update(choice_props)
                                      # Convert each property definition from JSON Structure to JSON Schema
                                    converted_properties = {}
                                    for prop_name, prop_def in choice_properties.items():
                                        converted_properties[prop_name] = self._convert_type_definition(prop_def)
                                    
                                    # Add selector constraint if this is an inline choice with selector
                                    if 'selector' in type_def:
                                        selector_field = type_def['selector']
                                        if selector_field in converted_properties:
                                            # Constrain the selector field to the specific choice name
                                            converted_properties[selector_field] = {'const': choice_name}
                                    
                                    merged_choice = {
                                        'type': 'object',
                                        'properties': converted_properties
                                    }
                                    
                                    # Add required fields from both base and choice
                                    required_fields = set()
                                    if 'required' in resolved_base and isinstance(resolved_base['required'], list):
                                        required_fields.update(resolved_base['required'])
                                    if 'required' in resolved_choice and isinstance(resolved_choice['required'], list):
                                        required_fields.update(resolved_choice['required'])
                                    if required_fields:
                                        merged_choice['required'] = sorted(list(required_fields))
                                        
                                    result['oneOf'].append(merged_choice)
                                else:
                                    result['oneOf'].append(converted_choice)
                            else:
                                result['oneOf'].append(converted_choice)
                        else:
                            # No inheritance, just use the converted choice
                            result['oneOf'].append(converted_choice)
            else:
                # Legacy format or error
                result['oneOf'] = []
                for choice in choices:
                    result['oneOf'].append(self._convert_type_definition(choice))
        
        return result

    def _convert_type_union(self, type_list: List[str], type_def: Dict[str, Any]) -> Dict[str, Any]:
        """Convert a JSON Structure type union to JSON Schema."""
        if len(type_list) == 1:
            # Single type, convert directly
            single_type_def = {'type': type_list[0]}
            single_type_def.update({k: v for k, v in type_def.items() if k != 'type'})
            return self._convert_type_definition(single_type_def)
        
        # Multiple types - create anyOf
        result = {
            'anyOf': []
        }
        
        for type_name in type_list:
            single_type_def = {'type': type_name}
            # Don't inherit other properties for union members
            converted = self.convert_type(type_name)
            result['anyOf'].append(converted)
        
        # Add any validation constraints to the union itself
        self._add_validation_constraints(result, type_def)
        
        return result

    def _add_validation_constraints(self, result: Dict[str, Any], type_def: Dict[str, Any]) -> None:
        """Add validation constraints from JSON Structure to JSON Schema."""        # String constraints
        if 'minLength' in type_def:
            result['minLength'] = type_def['minLength']
        if 'maxLength' in type_def:
            result['maxLength'] = type_def['maxLength']
        if 'pattern' in type_def:
            result['pattern'] = type_def['pattern']
        if 'format' in type_def:
            result['format'] = type_def['format']
        if 'enum' in type_def:
            result['enum'] = type_def['enum']
        if 'const' in type_def:
            result['const'] = type_def['const']
            
        # Numeric constraints
        if 'minimum' in type_def:
            result['minimum'] = type_def['minimum']
        if 'maximum' in type_def:
            result['maximum'] = type_def['maximum']
        if 'exclusiveMinimum' in type_def:
            result['exclusiveMinimum'] = type_def['exclusiveMinimum']
        if 'exclusiveMaximum' in type_def:
            result['exclusiveMaximum'] = type_def['exclusiveMaximum']
        if 'multipleOf' in type_def:
            result['multipleOf'] = type_def['multipleOf']
            
        # For decimal types, handle precision/scale
        if 'precision' in type_def or 'scale' in type_def:
            # Add custom keywords for precision/scale as JSON Schema doesn't have direct equivalents
            if 'precision' in type_def:
                result['x-precision'] = type_def['precision']
            if 'scale' in type_def:
                result['x-scale'] = type_def['scale']
        
        # Handle units and currencies as custom properties
        if 'unit' in type_def:
            result['x-unit'] = type_def['unit']
        if 'currency' in type_def:
            result['x-currency'] = type_def['currency']
            
        # Handle alternate names
        if 'altnames' in type_def:
            result['x-altnames'] = type_def['altnames']

    def _resolve_reference(self, ref_path: str) -> Optional[Dict[str, Any]]:
        """
        Resolve a $ref path to the actual type definition.
        
        Args:
            ref_path (str): The reference path (e.g., "#/definitions/Person")
            
        Returns:
            Optional[Dict[str, Any]]: The resolved type definition or None if not found
        """
        if not self.structure_document or not ref_path.startswith('#/'):
            return None
            
        # Remove the '#/' prefix and split the path
        path_parts = ref_path[2:].split('/')
        
        current = self.structure_document
        for part in path_parts:
            if isinstance(current, dict) and part in current:
                current = current[part]
            else:
                return None
                
        return current if isinstance(current, dict) else None

    def _resolve_type_with_inheritance(self, type_def: Dict[str, Any]) -> Dict[str, Any]:
        """
        Resolve a type definition, handling $extends inheritance.
        
        Args:
            type_def (Dict[str, Any]): The type definition that may have $extends
            
        Returns:
            Dict[str, Any]: The resolved type definition with inheritance applied
        """
        if '$extends' not in type_def:
            return type_def
            
        # Resolve the base type
        base_ref = type_def['$extends']
        base_type = self._resolve_reference(base_ref)
        
        if not base_type:
            # If we can't resolve the base, return the original type
            return type_def
            
        # Recursively resolve the base type's inheritance
        resolved_base = self._resolve_type_with_inheritance(base_type)
        
        # Merge the base type with the current type
        # The current type's properties override the base type's properties
        merged = {}
        merged.update(resolved_base)
        merged.update(type_def)
        
        # Special handling for properties - merge them
        if 'properties' in resolved_base and 'properties' in type_def:
            merged_props = {}
            merged_props.update(resolved_base['properties'])
            merged_props.update(type_def['properties'])
            merged['properties'] = merged_props
            
        # Remove $extends from the final result as it's not part of JSON Schema
        if '$extends' in merged:
            del merged['$extends']
            
        return merged


def convert_structure_to_json_schema(
    structure_file_path: str, 
    json_schema_path: str
) -> None:
    """
    Convert a JSON Structure file to JSON Schema format.
    
    Args:
        structure_file_path (str): Path to the input JSON Structure file
        json_schema_path (str): Path to the output JSON Schema file
    """
    # Read the JSON Structure file
    with open(structure_file_path, 'r', encoding='utf-8') as f:
        structure_content = f.read()
    
    # Convert to JSON Schema
    result = convert_structure_to_json_schema_string(structure_content)
    
    # Write the result
    with open(json_schema_path, 'w', encoding='utf-8') as f:
        f.write(result)


def convert_structure_to_json_schema_string(structure_content: str) -> str:
    """
    Convert a JSON Structure string to JSON Schema format.
    
    Args:
        structure_content (str): The JSON Structure document as a string
        
    Returns:
        str: The converted JSON Schema document as a string
    """
    converter = StructureToJsonConverter()
    
    try:
        structure_schema = json.loads(structure_content)
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON Structure document: {e}") from e
    
    # Convert the JSON Structure to JSON Schema
    result = converter.convert_structure_schema(structure_schema)
    
    return json.dumps(result, indent=2)



import json
import os
from typing import Dict, List, Union
from avrotize.common import is_generic_avro_type, render_template, pascal, camel, snake

INDENT = '    '

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None


class AvroToRust:
    """Converts Avro schema to Rust structs, including Serde and Avro marshalling methods"""

    def __init__(self, base_package: str = '') -> None:
        self.base_package = base_package.replace('.', '/').lower()
        self.output_dir = os.getcwd()
        self.generated_types_avro_namespace: Dict[str, str] = {}
        self.generated_types_rust_package: Dict[str, str] = {}
        self.avro_annotation = False
        self.serde_annotation = False
        
    reserved_words = [
            'as', 'break', 'const', 'continue', 'crate', 'else', 'enum', 'extern', 'false', 'fn', 'for', 'if', 'impl',
            'in', 'let', 'loop', 'match', 'mod', 'move', 'mut', 'pub', 'ref', 'return', 'self', 'Self', 'static',
            'struct', 'super', 'trait', 'true', 'type', 'unsafe', 'use', 'where', 'while', 'async', 'await', 'dyn',
        ]

    def safe_identifier(self, name: str) -> str:
        """Converts a name to a safe Rust identifier"""
        if name in AvroToRust.reserved_words:
            return f"{name}_"
        return name
    
    def escaped_identifier(self, name: str) -> str:
        """Converts a name to a safe Rust identifier with a leading r# prefix"""
        if name != "crate" and name in AvroToRust.reserved_words:
            return f"r#{name}"
        return name
    
    def safe_package(self, package: str) -> str:
        """Converts a package name to a safe Rust package name"""
        elements = package.split('::')
        return '::'.join([self.escaped_identifier(element) for element in elements])

    def map_primitive_to_rust(self, avro_fullname: str, is_optional: bool) -> str:
        """Maps Avro primitive types to Rust types"""
        optional_mapping = {
            'null': 'None',
            'boolean': 'Option<bool>',
            'int': 'Option<i32>',
            'long': 'Option<i64>',
            'float': 'Option<f32>',
            'double': 'Option<f64>',
            'bytes': 'Option<Vec<u8>>',
            'string': 'Option<String>',
        }
        required_mapping = {
            'null': 'None',
            'boolean': 'bool',
            'int': 'i32',
            'long': 'i64',
            'float': 'f32',
            'double': 'f64',
            'bytes': 'Vec<u8>',
            'string': 'String',
        }
        rust_fullname = avro_fullname
        if '.' in rust_fullname:
            type_name = pascal(avro_fullname.split('.')[-1])
            package_name = '::'.join(avro_fullname.split('.')[:-1]).lower()
            rust_fullname = self.safe_package(self.concat_package(package_name, type_name))
        if rust_fullname in self.generated_types_rust_package:
            return rust_fullname
        else:
            return required_mapping.get(avro_fullname, avro_fullname) if not is_optional else optional_mapping.get(avro_fullname, avro_fullname)

    def concat_package(self, package: str, name: str) -> str:
        """Concatenates package and name using a double colon separator"""
        return f"crate::{package.lower()}::{name.lower()}::{name}" if package else name

    def convert_avro_type_to_rust(self, field_name: str, avro_type: Union[str, Dict, List], namespace: str, nullable: bool = False) -> str:
        """Converts Avro type to Rust type"""
        ns = namespace.replace('.', '::').lower()
        type_name = ''
        if isinstance(avro_type, str):
            type_name = self.map_primitive_to_rust(avro_type, nullable)
        elif isinstance(avro_type, list):
            if is_generic_avro_type(avro_type):
                return 'serde_json::Value' if self.serde_annotation else 'std::collections::HashMap<String, String>'
            non_null_types = [t for t in avro_type if t != 'null']
            if len(non_null_types) == 1:
                # Rust apache-avro has a bug in the union type handling, so we need to swap the types
                # if the first type is not null
                if avro_type[0] != 'null':
                    avro_type[1] = avro_type[0]
                    avro_type[0] = 'null'
                if isinstance(non_null_types[0], str):
                    type_name = self.map_primitive_to_rust(non_null_types[0], True)
                else:
                    type_name = self.convert_avro_type_to_rust(field_name, non_null_types[0], namespace)
            else:
                type_name = self.generate_union_enum(field_name, avro_type, namespace)
        elif isinstance(avro_type, dict):
            if avro_type['type'] in ['record', 'enum']:
                type_name = self.generate_class_or_enum(avro_type, namespace)
            elif avro_type['type'] == 'fixed' or avro_type['type'] == 'bytes' and 'logicalType' in avro_type:
                if avro_type['logicalType'] == 'decimal':
                    return 'f64'
            elif avro_type['type'] == 'array':
                item_type = self.convert_avro_type_to_rust(field_name, avro_type['items'], namespace)
                return f"Vec<{item_type}>"
            elif avro_type['type'] == 'map':
                values_type = self.convert_avro_type_to_rust(field_name, avro_type['values'], namespace)
                return f"std::collections::HashMap<String, {values_type}>"
            elif 'logicalType' in avro_type:
                if avro_type['logicalType'] == 'date':
                    return 'chrono::NaiveDate'
                elif avro_type['logicalType'] == 'time-millis' or avro_type['logicalType'] == 'time-micros':
                    return 'chrono::NaiveTime'
                elif avro_type['logicalType'] == 'timestamp-millis' or avro_type['logicalType'] == 'timestamp-micros':
                    return 'chrono::NaiveDateTime'
                elif avro_type['logicalType'] == 'uuid':
                    return 'uuid::Uuid'
            else:
                type_name = self.convert_avro_type_to_rust(field_name, avro_type['type'], namespace)
        if type_name:
            return type_name
        return 'serde_json::Value' if self.serde_annotation else 'std::collections::HashMap<String, String>'

    def generate_class_or_enum(self, avro_schema: Dict, parent_namespace: str = '') -> str:
        """Generates a Rust struct or enum from an Avro schema"""
        namespace = avro_schema.get('namespace', parent_namespace).lower()
        if avro_schema['type'] == 'record':
            return self.generate_struct(avro_schema, namespace)
        elif avro_schema['type'] == 'enum':
            return self.generate_enum(avro_schema, namespace)
        return 'serde_json::Value'

    def generate_struct(self, avro_schema: Dict, parent_namespace: str) -> str:
        """Generates a Rust struct from an Avro record schema"""
        fields = []
        for field in avro_schema.get('fields', []):
            original_field_name = field['name']
            field_name = self.safe_identifier(snake(original_field_name))
            field_type = self.convert_avro_type_to_rust(field_name, field['type'], parent_namespace)
            serde_rename = field_name != original_field_name
            fields.append({
                'original_name': original_field_name,
                'name': field_name,
                'type': field_type,
                'serde_rename': serde_rename,
                'random_value': self.generate_random_value(field_type)
            })
        
        struct_name = self.safe_identifier(pascal(avro_schema['name']))
        ns = parent_namespace.replace('.', '::').lower()
        qualified_struct_name = self.safe_package(self.concat_package(ns, struct_name))
        if not 'namespace' in avro_schema:
            avro_schema['namespace'] = parent_namespace
        avro_schema_str = json.dumps(avro_schema)        
        avro_schema_str = avro_schema_str.replace('"', '§')
        avro_schema_str = f"\",\n{INDENT*2}\"".join(
            [avro_schema_str[i:i+80] for i in range(0, len(avro_schema_str), 80)])
        avro_schema_str = avro_schema_str.replace('§', '\\"')
        avro_schema_str = f"concat!(\"{avro_schema_str}\")"

        context = {
            'avro_annotation': self.avro_annotation,
            'serde_annotation': self.serde_annotation,
            'doc': avro_schema.get('doc', ''),
            'struct_name': struct_name,
            'fields': fields,
            'avro_schema': avro_schema_str,
            'json_match_predicates': [self.get_is_json_match_clause(f['original_name'], f['type']) for f in fields]
        }

        file_name = self.to_file_name(qualified_struct_name)
        target_file = os.path.join(self.output_dir, "src", file_name + ".rs")
        render_template('avrotorust/dataclass_struct.rs.jinja', target_file, **context)
        self.write_mod_rs(parent_namespace)

        self.generated_types_avro_namespace[qualified_struct_name] = "struct"
        self.generated_types_rust_package[qualified_struct_name] = "struct"

        return qualified_struct_name
    
    def get_is_json_match_clause(self, field_name: str, field_type: str, for_union=False) -> str:
        """Generates the is_json_match clause for a field"""
        ref = f'node[\"{field_name}\"]' if not for_union else 'node'
        if field_type == 'String' or field_type == 'Option<String>':
            return f"{ref}.is_string()"
        elif field_type == 'bool' or field_type == 'Option<bool>':
            return f"{ref}.is_boolean()"
        elif field_type == 'i32' or field_type == 'Option<i32>':
            return f"{ref}.is_i64()"
        elif field_type == 'i64' or field_type == 'Option<i64>':
            return f"{ref}.is_i64()"
        elif field_type == 'f32' or field_type == 'Option<f32>':
            return f"{ref}.is_f64()"
        elif field_type == 'f64' or field_type == 'Option<f64>':
            return f"{ref}.is_f64()"
        elif field_type == 'Vec<u8>' or field_type == 'Option<Vec<u8>>':
            return f"{ref}.is_array()"
        elif field_type == 'serde_json::Value' or field_type == 'std::collections::HashMap<String, String>':
            return f"{ref}.is_object()"
        elif field_type.startswith('std::collections::HashMap<String, '):
            return f"{ref}.is_object()"
        elif field_type.startswith('Vec<'):
            return f"{ref}.is_array()"
        else:
            return f"{field_type}::is_json_match(&{ref})"


    def generate_enum(self, avro_schema: Dict, parent_namespace: str) -> str:
        """Generates a Rust enum from an Avro enum schema"""
        symbols = avro_schema.get('symbols', [])
        enum_name = self.safe_identifier(pascal(avro_schema['name']))
        ns = parent_namespace.replace('.', '::').lower()
        qualified_enum_name = self.safe_package(self.concat_package(ns, enum_name))
        
        if not 'namespace' in avro_schema:
            avro_schema['namespace'] = parent_namespace
        avro_schema_str = json.dumps(avro_schema)
        avro_schema_str = avro_schema_str.replace('"', '§')
        avro_schema_str = f"\",\n{INDENT*2}\"".join(
            [avro_schema_str[i:i+80] for i in range(0, len(avro_schema_str), 80)])
        avro_schema_str = avro_schema_str.replace('§', '\\"')
        avro_schema_str = f"concat!(\"{avro_schema_str}\")"

        context = {
            'avro_annotation': self.avro_annotation,
            'serde_annotation': self.serde_annotation,
            'enum_name': enum_name,
            'symbols': symbols,
            'avro_schema': avro_schema_str,
        }

        file_name = self.to_file_name(qualified_enum_name)
        target_file = os.path.join(self.output_dir, "src", file_name + ".rs")
        render_template('avrotorust/dataclass_enum.rs.jinja', target_file, **context)
        self.write_mod_rs(parent_namespace)

        self.generated_types_avro_namespace[qualified_enum_name] = "enum"
        self.generated_types_rust_package[qualified_enum_name] = "enum"

        return qualified_enum_name

    def generate_union_enum(self, field_name: str, avro_type: List, namespace: str) -> str:
        """Generates a union enum for Rust"""
        ns = namespace.replace('.', '::').lower()
        union_enum_name = pascal(field_name) + 'Union'
        union_types = [self.convert_avro_type_to_rust(field_name + "Option" + str(i), t, namespace) for i, t in enumerate(avro_type) if t != 'null']
        union_fields = [
            {
                'name': pascal(t.rsplit('::',1)[-1]), 
                'type': t, 
                'random_value': self.generate_random_value(t),
                'default_value': 'Default::default()',
                'json_match_predicate': self.get_is_json_match_clause(field_name, t, for_union=True),
            } for i, t in enumerate(union_types)]
        qualified_union_enum_name = self.safe_package(self.concat_package(ns, union_enum_name))
        context = {
            'serde_annotation': self.serde_annotation,
            'union_enum_name': union_enum_name,
            'union_fields': union_fields,
            'json_match_predicates': [self.get_is_json_match_clause(f['name'], f['type'], for_union=True) for f in union_fields]
        }

        file_name = self.to_file_name(qualified_union_enum_name)
        target_file = os.path.join(self.output_dir, "src", file_name + ".rs").lower()
        render_template('avrotorust/dataclass_union.rs.jinja', target_file, **context)
        self.generated_types_avro_namespace[qualified_union_enum_name] = "union"
        self.generated_types_rust_package[qualified_union_enum_name] = "union"
        self.write_mod_rs(namespace)

        return qualified_union_enum_name

    def to_file_name(self, qualified_name):
        """Converts a qualified union enum name to a file name"""
        if qualified_name.startswith('crate::'):
            qualified_name = qualified_name[(len('crate::')):]
        qualified_name = qualified_name.replace('r#', '')
        return qualified_name.rsplit('::',1)[0].replace('::', os.sep).lower()
    
    def generate_random_value(self, rust_type: str) -> str:
        """Generates a random value for a given Rust type"""
        if rust_type == 'String' or rust_type == 'Option<String>':
            return 'format!("random_string_{}", rand::Rng::gen::<u32>(&mut rng))'
        elif rust_type == 'bool' or rust_type == 'Option<bool>':
            return 'rand::Rng::gen::<bool>(&mut rng)'
        elif rust_type == 'i32' or rust_type == 'Option<i32>':
            return 'rand::Rng::gen_range(&mut rng, 0..100)'
        elif rust_type == 'i64' or rust_type == 'Option<i64>':
            return 'rand::Rng::gen_range(&mut rng, 0..100) as i64'
        elif rust_type == 'f32' or rust_type == 'Option<f32>':
            return '(rand::Rng::gen::<f32>(&mut rng)*1000.0).round()/1000.0'
        elif rust_type == 'f64' or rust_type == 'Option<f64>':
            return '(rand::Rng::gen::<f64>(&mut rng)*1000.0).round()/1000.0'
        elif rust_type == 'Vec<u8>' or rust_type == 'Option<Vec<u8>>':
            return 'vec![rand::Rng::gen::<u8>(&mut rng); 10]'
        elif rust_type == 'chrono::NaiveDate':
            return 'chrono::NaiveDate::from_ymd(rand::Rng::gen_range(&mut rng, 2000..2023), rand::Rng::gen_range(&mut rng, 1..13), rand::Rng::gen_range(&mut rng, 1..29))'
        elif rust_type == 'chrono::NaiveTime':
            return 'chrono::NaiveTime::from_hms(rand::Rng::gen_range(&mut rng, 0..24),rand::Rng::gen_range(&mut rng, 0..60), rand::Rng::gen_range(&mut rng, 0..60))'
        elif rust_type == 'chrono::NaiveDateTime':
            return 'chrono::NaiveDateTime::new(chrono::NaiveDate::from_ymd(rand::Rng::gen_range(&mut rng, 2000..2023), rand::Rng::gen_range(&mut rng, 1..13), rand::Rng::gen_range(&mut rng, 1..29)), chrono::NaiveTime::from_hms(rand::Rng::gen_range(&mut rng, 0..24), rand::Rng::gen_range(&mut rng, 0..60), rand::Rng::gen_range(&mut rng, 0..60)))'
        elif rust_type == 'uuid::Uuid':
            return 'uuid::Uuid::new_v4()'
        elif rust_type.startswith('std::collections::HashMap<String, '):
            inner_type = rust_type.split(', ')[1][:-1]
            return f'(0..3).map(|_| (format!("key_{{}}", rand::Rng::gen::<u32>(&mut rng)), {self.generate_random_value(inner_type)})).collect()'
        elif rust_type.startswith('Vec<'):
            inner_type = rust_type[4:-1]
            return f'(0..3).map(|_| {self.generate_random_value(inner_type)}).collect()'
        elif rust_type in self.generated_types_rust_package:
            return f'{rust_type}::generate_random_instance()'
        else:
            return 'Default::default()'

    def write_mod_rs(self, namespace: str):
        """Writes the mod.rs file for a Rust module"""
        directories = namespace.split('.')
        for i in range(len(directories)):
            sub_package = '::'.join(directories[:i + 1])
            directory_path = os.path.join(
                self.output_dir, "src", sub_package.replace('.', os.sep).replace('::', os.sep))
            if not os.path.exists(directory_path):
                os.makedirs(directory_path, exist_ok=True)
            mod_rs_path = os.path.join(directory_path, "mod.rs")
            
            types = [file.replace('.rs', '') for file in os.listdir(directory_path) if file.endswith('.rs') and file != "mod.rs"]
            mod_statements = '\n'.join(f'pub mod {self.escaped_identifier(typ.lower())};' for typ in types)
            mods = [dir for dir in os.listdir(directory_path) if os.path.isdir(os.path.join(directory_path, dir))]
            mod_statements += '\n' + '\n'.join(f'pub mod {self.escaped_identifier(mod.lower())};' for mod in mods)

            with open(mod_rs_path, 'w', encoding='utf-8') as file:
                file.write(mod_statements)

    def write_cargo_toml(self):
        """Writes the Cargo.toml file for the Rust project"""
        dependencies = []
        if self.serde_annotation or self.avro_annotation:
            dependencies.append('serde = { version = "1.0", features = ["derive"] }')
            dependencies.append('serde_json = "1.0"')
        dependencies.append('chrono = { version = "0.4", features = ["serde"] }')
        dependencies.append('uuid = { version = "1.11", features = ["serde", "v4"] }')
        if self.avro_annotation or self.serde_annotation:
            dependencies.append('flate2 = "1.0"')
        if self.avro_annotation:
            dependencies.append('apache-avro = "0.17"')
            dependencies.append('lazy_static = "1.4"')
        dependencies.append('rand = "0.8"')

        cargo_toml_content =  f"[package]\n"
        cargo_toml_content += f"name = \"{self.base_package.replace('/', '_')}\"\n"
        cargo_toml_content += f"version = \"0.1.0\"\n"
        cargo_toml_content += f"edition = \"2021\"\n\n"
        cargo_toml_content += f"[dependencies]\n"
        cargo_toml_content += "\n".join(f"{dependency}" for dependency in dependencies)
        cargo_toml_path = os.path.join(self.output_dir, "Cargo.toml")
        with open(cargo_toml_path, 'w', encoding='utf-8') as file:
            file.write(cargo_toml_content)

    def write_lib_rs(self):
        """Writes the lib.rs file for the Rust project"""
        modules = {name[(len('crate::')):].split('::')[0] for name in self.generated_types_rust_package}
        mod_statements = '\n'.join(f'pub mod {module};' for module in modules)
        
        lib_rs_content = f"""
// This is the library entry point

{mod_statements}
"""
        lib_rs_path = os.path.join(self.output_dir, "src", "lib.rs")
        if not os.path.exists(os.path.dirname(lib_rs_path)):
            os.makedirs(os.path.dirname(lib_rs_path), exist_ok=True)
        with open(lib_rs_path, 'w', encoding='utf-8') as file:
            file.write(lib_rs_content)

    def convert_schema(self, schema: JsonNode, output_dir: str):
        """Converts Avro schema to Rust"""
        if not isinstance(schema, list):
            schema = [schema]
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        self.output_dir = output_dir
        for avro_schema in (x for x in schema if isinstance(x, dict)):
            self.generate_class_or_enum(avro_schema)

        self.write_cargo_toml()
        self.write_lib_rs()

    def convert(self, avro_schema_path: str, output_dir: str):
        """Converts Avro schema to Rust"""
        with open(avro_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        self.convert_schema(schema, output_dir)


def convert_avro_to_rust(avro_schema_path, rust_file_path, package_name='', avro_annotation=False, serde_annotation=False):
    """Converts Avro schema to Rust structs

    Args:
        avro_schema_path (str): Avro input schema path  
        rust_file_path (str): Output Rust file path 
        package_name (str): Base package name
        avro_annotation (bool): Include Avro annotations
        serde_annotation (bool): Include Serde annotations
    """
    
    if not package_name:
        package_name = os.path.splitext(os.path.basename(avro_schema_path))[0].lower().replace('-', '_')
        
    avrotorust = AvroToRust()
    avrotorust.base_package = package_name
    avrotorust.avro_annotation = avro_annotation
    avrotorust.serde_annotation = serde_annotation
    avrotorust.convert(avro_schema_path, rust_file_path)


def convert_avro_schema_to_rust(avro_schema: JsonNode, output_dir: str, package_name='', avro_annotation=False, serde_annotation=False):
    """Converts Avro schema to Rust structs

    Args:
        avro_schema (JsonNode): Avro schema as a dictionary or list of dictionaries
        output_dir (str): Output directory path 
        package_name (str): Base package name
        avro_annotation (bool): Include Avro annotations
        serde_annotation (bool): Include Serde annotations
    """
    avrotorust = AvroToRust()
    avrotorust.base_package = package_name
    avrotorust.avro_annotation = avro_annotation
    avrotorust.serde_annotation = serde_annotation
    avrotorust.convert_schema(avro_schema, output_dir)



# pylint: disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements, line-too-long

""" Generates Java classes from Avro schema """
import json
import os
from typing import Dict, List, Tuple, Union
from avrotize.constants import AVRO_VERSION, JACKSON_VERSION, JDK_VERSION

from avrotize.common import pascal, camel, is_generic_avro_type

INDENT = '    '
POM_CONTENT = """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>
    <groupId>{groupid}</groupId>
    <artifactId>{artifactid}</artifactId>
    <version>1.0-SNAPSHOT</version>
    <properties>
        <maven.compiler.source>{JDK_VERSION}</maven.compiler.source>
        <maven.compiler.target>{JDK_VERSION}</maven.compiler.target>
    </properties>
    <dependencies>
        <dependency>
            <groupId>org.apache.avro</groupId>
            <artifactId>avro</artifactId>
            <version>{AVRO_VERSION}</version>
        </dependency>
        <dependency>
            <groupId>com.fasterxml.jackson</groupId>
            <artifactId>jackson-bom</artifactId>
            <version>{JACKSON_VERSION}</version>
            <type>pom</type>
        </dependency>
    </dependencies>
</project>
"""

PREAMBLE_TOBYTEARRAY = \
"""
byte[] result = null;
String mediaType = contentType.split(";")[0].trim().toLowerCase();
"""


EPILOGUE_TOBYTEARRAY_COMPRESSION = \
    """
if (result != null && mediaType.endsWith("+gzip")) {
    try (ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
         GZIPOutputStream gzipOutputStream = new GZIPOutputStream(byteArrayOutputStream)) {
        gzipOutputStream.write(result);
        gzipOutputStream.finish();
        result = byteArrayOutputStream.toByteArray();
    } catch (IOException e) {
        throw new UnsupportedOperationException("Error compressing data to gzip");
    }
}
"""

EPILOGUE_TOBYTEARRAY = \
"""
throw new UnsupportedOperationException("Unsupported media type + mediaType");
"""

PREAMBLE_FROMDATA_COMPRESSION = \
"""
if (mediaType.endsWith("+gzip")) {
    InputStream stream = null;
    
    if (data instanceof InputStream) {
        stream = (InputStream) data;
    } else if (data instanceof byte[]) {
        stream = new ByteArrayInputStream((byte[]) data);
    } else {
        throw new UnsupportedOperationException("Data is not of a supported type for gzip decompression");
    }
    
    try (InputStream gzipStream = new GZIPInputStream(stream);
         ByteArrayOutputStream outputStream = new ByteArrayOutputStream()) {
        byte[] buffer = new byte[1024];
        int bytesRead;
        while ((bytesRead = gzipStream.read(buffer)) != -1) {
            outputStream.write(buffer, 0, bytesRead);
        }
        data = outputStream.toByteArray();
    } catch (IOException e) {
        e.printStackTrace();
    }
}
"""


JSON_FROMDATA_THROWS = \
    ",JsonProcessingException, IOException"
JSON_FROMDATA = \
    """
if ( mediaType == "application/json") {
    if (data instanceof byte[]) {
        ByteArrayInputStream stream = new ByteArrayInputStream((byte[]) data);
        return (new ObjectMapper()).readValue(stream, {typeName}.class);
    }
    else if (data instanceof InputStream) {
        return (new ObjectMapper()).readValue((InputStream)data, {typeName}.class);
    }
    else if (data instanceof JsonNode) {
        return (new ObjectMapper()).readValue(((JsonNode)data).toString(), {typeName}.class);
    }
    else if ( data instanceof String) {
        return (new ObjectMapper()).readValue(((String)data), {typeName}.class);
    }
    throw new UnsupportedOperationException("Data is not of a supported type for JSON conversion to {typeName}");
}
"""
JSON_TOBYTEARRAY_THROWS = ",JsonProcessingException"
JSON_TOBYTEARRAY = \
    """
if ( mediaType == "application/json") {    
    result = new ObjectMapper().writeValueAsBytes(this);
}
"""

AVRO_FROMDATA_THROWS = ",IOException"
AVRO_FROMDATA = \
    """
if ( mediaType == "avro/binary" || mediaType == "application/vnd.apache.avro+avro") {
    if (data instanceof byte[]) {
        return AVROREADER.read(new {typeName}(), DecoderFactory.get().binaryDecoder((byte[])data, null));
    } else if (data instanceof InputStream) {
        return AVROREADER.read(new {typeName}(), DecoderFactory.get().binaryDecoder((InputStream)data, null));
    }
    throw new UnsupportedOperationException("Data is not of a supported type for Avro conversion to {typeName}");
} else if ( mediaType == "avro/json" || mediaType == "application/vnd.apache.avro+json") {
    if (data instanceof byte[]) {
        return AVROREADER.read(new {typeName}(), DecoderFactory.get().jsonDecoder({typeName}.AVROSCHEMA, new ByteArrayInputStream((byte[])data)));
    } else if (data instanceof InputStream) {
        return AVROREADER.read(new {typeName}(), DecoderFactory.get().jsonDecoder({typeName}.AVROSCHEMA, (InputStream)data));
    } else if (data instanceof String) {
        return AVROREADER.read(new {typeName}(), DecoderFactory.get().jsonDecoder({typeName}.AVROSCHEMA, (String)data));
    }
    throw new UnsupportedOperationException("Data is not of a supported type for Avro conversion to {typeName}");
}
"""


AVRO_TOBYTEARRAY_THROWS = ",IOException"
AVRO_TOBYTEARRAY = \
    """
if ( mediaType == "avro/binary" || mediaType == "application/vnd.apache.avro+avro") {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    Encoder encoder = EncoderFactory.get().binaryEncoder(out, null);
    AVROWRITER.write(this, encoder);
    encoder.flush();
    result = out.toByteArray();
}
else if ( mediaType == "avro/json" || mediaType == "application/vnd.apache.avro+json") {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    Encoder encoder = EncoderFactory.get().jsonEncoder({typeName}.AVROSCHEMA, out);
    AVROWRITER.write(this, encoder);
    encoder.flush();
    result = out.toByteArray();
}
"""


JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None


def flatten_type_name(name: str) -> str:
    """Strips the namespace from a name"""
    if name.endswith('[]'):
        return flatten_type_name(name[:-2]+'Array')
    base_name = pascal(name.replace(' ', '').split('.')[-1].replace('>', '').replace('<', '').replace(',', ''))
    return base_name


def is_java_reserved_word(word: str) -> bool:
    """Checks if a word is a Java reserved word"""
    reserved_words = [
        'abstract', 'assert', 'boolean', 'break', 'byte', 'case', 'catch', 'char', 'class', 'const',
        'continue', 'default', 'do', 'double', 'else', 'enum', 'extends', 'final', 'finally', 'float',
        'for', 'goto', 'if', 'implements', 'import', 'instanceof', 'int', 'interface', 'long', 'native',
        'new', 'package', 'private', 'protected', 'public', 'return', 'short', 'static', 'strictfp',
        'super', 'switch', 'synchronized', 'this', 'throw', 'throws', 'transient', 'try', 'void', 'volatile',
        'while', 'true', 'false', 'null', 'record', 
    ]
    return word in reserved_words


class AvroToJava:
    """Converts Avro schema to Java classes, including Jackson annotations and Avro SpecificRecord methods"""

    def __init__(self, base_package: str = '') -> None:
        self.base_package = base_package.replace('.', '/')
        self.output_dir = os.getcwd()
        self.avro_annotation = False
        self.jackson_annotations = False
        self.pascal_properties = False
        self.generated_types_avro_namespace: Dict[str,str] = {}
        self.generated_types_java_package: Dict[str,str] = {}

    def qualified_name(self, package: str, name: str) -> str:
        """Concatenates package and name using a dot separator"""
        slash_package_name = package.replace('.', '/')
        safe_package_slash = self.safe_package(slash_package_name.lower())
        safe_package = safe_package_slash.replace('/', '.')
        return f"{safe_package}.{name}" if package else name
    
    def join_packages(self, parent_package: str, package: str) -> str:
        """Joins package and name using a dot separator"""
        if parent_package and package:
            return f"{parent_package}.{package}".lower()
        elif parent_package:
            return parent_package.lower()
        return package.lower()

    class JavaType:
        """Java type definition"""

        def __init__(self, type_name: str, union_types: List['AvroToJava.JavaType'] | None = None, is_class: bool = False, is_enum: bool = False) -> None:
            self.type_name = type_name
            self.union_types = union_types
            self.is_class = is_class
            self.is_enum = is_enum

    def safe_identifier(self, name: str, class_name: str = '') -> str:
        """Converts a name to a safe Java identifier"""
        if is_java_reserved_word(name):
            return f"_{name}"
        if class_name and name == class_name:
            return f"{name}_"
        return name

    def safe_package(self, packageName: str) -> str:
        """Converts a name to a safe Java identifier by checking each path segment"""
        segments = packageName.split('/')
        safe_segments = [
            self.safe_identifier(segment)
            for segment in segments
        ]
        
        return '/'.join(safe_segments)
    
    def map_primitive_to_java(self, avro_type: str, is_optional: bool) -> JavaType:
        """Maps Avro primitive types to Java types"""
        optional_mapping = {
            'null': 'Void',
            'boolean': 'Boolean',
            'int': 'Integer',
            'long': 'Long',
            'float': 'Float',
            'double': 'Double',
            'bytes': 'byte[]',
            'string': 'String',
        }
        required_mapping = {
            'null': 'void',
            'boolean': 'boolean',
            'int': 'int',
            'long': 'long',
            'float': 'float',
            'double': 'double',
            'bytes': 'byte[]',
            'string': 'String',
        }
        if '.' in avro_type:
            type_name = avro_type.split('.')[-1]
            package_name = '.'.join(avro_type.split('.')[:-1]).lower()
            avro_type = self.qualified_name(package_name, type_name)
        if avro_type in self.generated_types_avro_namespace:
            kind = self.generated_types_avro_namespace[avro_type]
            qualified_class_name = self.qualified_name(self.base_package, avro_type)
            return AvroToJava.JavaType(qualified_class_name, is_class=kind=="class", is_enum=kind=="enum")
        else:
            return AvroToJava.JavaType(required_mapping.get(avro_type, avro_type) if not is_optional else optional_mapping.get(avro_type, avro_type))
    
    def is_java_primitive(self, java_type: JavaType) -> bool:
        """Checks if a Java type is a primitive type"""
        return java_type.type_name in [
            'void', 'boolean', 'int', 'long', 'float', 'double', 'byte[]', 'String',
            'Boolean', 'Integer', 'Long', 'Float', 'Double', 'Void']
        
    def is_java_optional_type(self, java_type: JavaType) -> bool:
        """Checks if a Java type is an optional type"""
        return java_type.type_name in ['Void', 'Boolean', 'Integer', 'Long', 'Float', 'Double']
        
    def is_java_numeric_type(self, java_type: JavaType) -> bool:
        """Checks if a Java type is a numeric type"""
        return java_type.type_name in ['int', 'long', 'float', 'double', 'Integer', 'Long', 'Float', 'Double']
    
    def is_java_integer_type(self, java_type: JavaType) -> bool:
        """Checks if a Java type is an integer type"""
        return java_type.type_name in ['int', 'long', 'Integer', 'Long']

    def convert_avro_type_to_java(self, class_name: str, field_name: str, avro_type: Union[str, Dict, List], parent_package: str, nullable: bool = False) -> JavaType:
        """Converts Avro type to Java type"""
        if isinstance(avro_type, str):
            return self.map_primitive_to_java(avro_type, nullable)
        elif isinstance(avro_type, list):
            if (is_generic_avro_type(avro_type)):
                return AvroToJava.JavaType('Object')
            non_null_types = [t for t in avro_type if t != 'null']
            if len(non_null_types) == 1:
                if isinstance(non_null_types[0], str):
                    return self.map_primitive_to_java(non_null_types[0], True)
                else:
                    return self.convert_avro_type_to_java(class_name, field_name, non_null_types[0], parent_package)
            else:
                if self.jackson_annotations:
                    return AvroToJava.JavaType(self.generate_embedded_union_class_jackson(class_name, field_name, non_null_types, parent_package, write_file=True), is_class=True)
                else:
                    types: List[AvroToJava.JavaType] = [self.convert_avro_type_to_java(
                        class_name, field_name, t, parent_package) for t in non_null_types]
                    return AvroToJava.JavaType('Object', types)
        elif isinstance(avro_type, dict):
            if avro_type['type'] in ['record', 'enum']:
                return self.generate_class_or_enum(avro_type, parent_package, write_file=True)
            elif avro_type['type'] == 'fixed':
                if 'logicalType' in avro_type and avro_type['logicalType'] == 'decimal':
                    return AvroToJava.JavaType('BigDecimal')
                return AvroToJava.JavaType('byte[]')
            elif avro_type['type'] == 'bytes' and 'logicalType' in avro_type:
                if avro_type['logicalType'] == 'decimal':
                    return AvroToJava.JavaType('BigDecimal')
            elif avro_type['type'] == 'array':
                item_type = self.convert_avro_type_to_java(class_name, field_name, avro_type['items'], parent_package, nullable=True).type_name
                return AvroToJava.JavaType(f"List<{item_type}>")
            elif avro_type['type'] == 'map':
                values_type = self.convert_avro_type_to_java(class_name, field_name, avro_type['values'], parent_package, nullable=True).type_name
                return AvroToJava.JavaType(f"Map<String,{values_type}>")
            elif 'logicalType' in avro_type:
                if avro_type['logicalType'] == 'date':
                    return AvroToJava.JavaType('LocalDate')
                elif avro_type['logicalType'] == 'time-millis' or avro_type['logicalType'] == 'time-micros':
                    return AvroToJava.JavaType('LocalTime')
                elif avro_type['logicalType'] == 'timestamp-millis' or avro_type['logicalType'] == 'timestamp-micros':
                    return AvroToJava.JavaType('Instant')
                elif avro_type['logicalType'] == 'local-timestamp-millis' or avro_type['logicalType'] == 'local-timestamp-micros':
                    return AvroToJava.JavaType('LocalDateTime')
                elif avro_type['logicalType'] == 'uuid':
                    return AvroToJava.JavaType('UUID')
                elif avro_type['logicalType'] == 'duration':
                    return AvroToJava.JavaType('Duration')
            return self.convert_avro_type_to_java(class_name, field_name, avro_type['type'], parent_package)
        return 'Object'

    def generate_class_or_enum(self, avro_schema: Dict, parent_package: str, write_file: bool = True) -> JavaType:
        """ Generates a Java class or enum from an Avro schema """
        if avro_schema['type'] == 'record':
            return self.generate_class(avro_schema, parent_package, write_file)
        elif avro_schema['type'] == 'enum':
            return self.generate_enum(avro_schema, parent_package, write_file)
        return AvroToJava.JavaType('Object')

    def generate_class(self, avro_schema: Dict, parent_package: str, write_file: bool) -> JavaType:
        """ Generates a Java class from an Avro record schema """
        class_definition = ''
        if 'doc' in avro_schema:
            class_definition += f"/** {avro_schema['doc']} */\n"
        namespace = avro_schema.get('namespace', parent_package)
        if not 'namespace' in avro_schema:
            avro_schema['namespace'] = namespace
        package = self.join_packages(self.base_package, namespace).replace('.', '/').lower()
        package = package.replace('.', '/').lower()
        package = self.safe_package(package)
        class_name = self.safe_identifier(avro_schema['name'])
        namespace_qualified_name = self.qualified_name(namespace,avro_schema['name'])
        qualified_class_name = self.qualified_name(package.replace('/', '.'), class_name)
        if namespace_qualified_name in self.generated_types_avro_namespace:
            return AvroToJava.JavaType(qualified_class_name, is_class=True)
        self.generated_types_avro_namespace[namespace_qualified_name] = "class"
        self.generated_types_java_package[qualified_class_name] = "class"
        fields_str = [self.generate_property(class_name, field, namespace) for field in avro_schema.get('fields', [])]
        class_body = "\n".join(fields_str)
        class_definition += f"public class {class_name}"
        if self.avro_annotation:
            class_definition += " implements SpecificRecord"
        class_definition += " {\n"
        class_definition += f"{INDENT}public {class_name}() {{}}\n"
        class_definition += class_body

        if self.avro_annotation:
            class_definition += f"\n{INDENT}public {class_name}(GenericData.Record record) {{\n"
            class_definition += f"{INDENT*2}for( int i = 0; i < record.getSchema().getFields().size(); i++ ) {{\n"
            class_definition += f"{INDENT*3}this.put(i, record.get(i));\n"
            class_definition += f"{INDENT*2}}}\n"
            class_definition += f"{INDENT}}}\n"

        if self.avro_annotation:
            avro_schema_json = json.dumps(avro_schema)
            avro_schema_json = avro_schema_json.replace('"', '§')
            avro_schema_json = f"\"+\n{INDENT}\"".join(
                [avro_schema_json[i:i+80] for i in range(0, len(avro_schema_json), 80)])
            avro_schema_json = avro_schema_json.replace('§', '\\"')
            class_definition += f"\n\n{INDENT}public static final Schema AVROSCHEMA = new Schema.Parser().parse(\n{INDENT}\"{avro_schema_json}\");"
            class_definition += f"\n{INDENT}public static final DatumWriter<{class_name}> AVROWRITER = new SpecificDatumWriter<{class_name}>(AVROSCHEMA);"
            class_definition += f"\n{INDENT}public static final DatumReader<{class_name}> AVROREADER = new SpecificDatumReader<{class_name}>(AVROSCHEMA);\n"

            if self.jackson_annotations:
                class_definition += f"\n{INDENT}@JsonIgnore"
            class_definition += f"\n{INDENT}@Override\n{INDENT}public Schema getSchema(){{ return AVROSCHEMA; }}\n"
            class_definition += self.generate_avro_get_method(class_name, avro_schema.get('fields', []), namespace)
            class_definition += self.generate_avro_put_method(class_name, avro_schema.get('fields', []), namespace)

        # emit toByteArray method
        class_definition += f"\n\n{INDENT}/**\n{INDENT} * Converts the object to a byte array\n{INDENT} * @param contentType the content type of the byte array\n{INDENT} * @return the byte array\n{INDENT} */\n"
        class_definition += f"{INDENT}public byte[] toByteArray(String contentType) throws UnsupportedOperationException" + \
            f"{ JSON_TOBYTEARRAY_THROWS if self.jackson_annotations else '' }" + \
            f"{ AVRO_TOBYTEARRAY_THROWS if self.avro_annotation else '' }  {{"
        if self.jackson_annotations or self.avro_annotation:
            class_definition += f'\n{INDENT*2}'.join((PREAMBLE_TOBYTEARRAY).split("\n"))
        if self.avro_annotation:
            class_definition += f'\n{INDENT*2}'+f'\n{INDENT*2}'.join(
                AVRO_TOBYTEARRAY.strip().replace("{typeName}", class_name).split("\n"))
        if self.jackson_annotations:
            class_definition += f'\n{INDENT*2}'+f'\n{INDENT*2}'.join(
                JSON_TOBYTEARRAY.strip().replace("{typeName}", class_name).split("\n"))
        if self.avro_annotation or self.jackson_annotations:
            class_definition += f'\n{INDENT*2}'.join(EPILOGUE_TOBYTEARRAY_COMPRESSION.split("\n"))
            class_definition += f'\n{INDENT*2}if ( result != null ) {{ return result; }}'        
        class_definition += (f'\n{INDENT*2}'.join((EPILOGUE_TOBYTEARRAY.strip()).split("\n")))+f"\n{INDENT}}}"

        # emit fromData factory method
        class_definition += f"\n\n{INDENT}/**\n{INDENT} * Converts the data to an object\n{INDENT} * @param data the data to convert\n{INDENT} * @param contentType the content type of the data\n{INDENT} * @return the object\n{INDENT} */\n"
        class_definition += f"{INDENT}public static {class_name} fromData(Object data, String contentType) throws UnsupportedOperationException" + \
            f"{ JSON_FROMDATA_THROWS if self.jackson_annotations else '' }" + \
            f"{ AVRO_FROMDATA_THROWS if self.avro_annotation else '' }  {{"
        class_definition += f'\n{INDENT*2}if ( data instanceof {class_name}) return ({class_name})data;'
        
        if self.avro_annotation or self.jackson_annotations:
            class_definition += f'\n{INDENT*2}String mediaType = contentType.split(";")[0].trim().toLowerCase();'
            class_definition += f'\n{INDENT*2}'.join((PREAMBLE_FROMDATA_COMPRESSION).split("\n"))
        if self.avro_annotation:
            class_definition += f'\n{INDENT*2}'+f'\n{INDENT*2}'.join(
                AVRO_FROMDATA.strip().replace("{typeName}", class_name).split("\n"))
        if self.jackson_annotations:
            class_definition += f'\n{INDENT*2}'+f'\n{INDENT*2}'.join(
                JSON_FROMDATA.strip().replace("{typeName}", class_name).split("\n"))
        class_definition += f"\n{INDENT*2}throw new UnsupportedOperationException(\"Unsupported media type \"+ contentType);\n{INDENT}}}"
        
        if self.jackson_annotations:
            class_definition += self.create_is_json_match_method(avro_schema, avro_schema.get('namespace', namespace), class_name)

        class_definition += "\n}"

        if write_file:
            self.write_to_file(package, class_name, class_definition)
        return AvroToJava.JavaType(qualified_class_name, is_class=True)
    
    def create_is_json_match_method(self, avro_schema, parent_namespace, class_name) -> str:
        """ Generates the isJsonMatch method for a class using Jackson """
        predicates = ''
        class_definition = ''
        class_definition += f"\n\n{INDENT}/**\n{INDENT} * Checks if the JSON node matches the schema\n{INDENT}"
        class_definition += f"\n{INDENT}@param node The JSON node to check */"
        class_definition += f"\n{INDENT}public static boolean isJsonMatch(com.fasterxml.jackson.databind.JsonNode node)\n{INDENT}{{"
        field_defs = ''
        
        field_count = 0
        for field in avro_schema.get('fields', []):
            if field_count > 0:
                field_defs += f" && \n{INDENT*3}"
            field_count += 1
            field_name = field['name']
            if field_name == class_name:
                field_name += "_"
            field_type = self.convert_avro_type_to_java(class_name, field_name, field['type'], parent_namespace)
            predicate, clause = self.get_is_json_match_clause(class_name, field_name, field_type)
            field_defs += clause
            if predicate:
                predicates += predicate + "\n"
        if ( len(predicates) > 0 ):
            class_definition += f'\n{INDENT*2}'+f'\n{INDENT*2}'.join(predicates.split('\n'))
        class_definition += f"\n{INDENT*2}return {field_defs}"
        class_definition += f";\n{INDENT}}}"
        return class_definition
    
    def get_is_json_match_clause(self, class_name: str, field_name: str, field_type: JavaType) -> Tuple[str, str]:
        """ Generates the isJsonMatch clause for a field using Jackson """
        class_definition = ''
        predicates = ''
        field_name_js = field_name
        is_optional = self.is_java_optional_type(field_type)        

        if is_optional:
            node_check = f"!node.has(\"{field_name_js}\") || node.get(\"{field_name_js}\").isNull() || node.get(\"{field_name_js}\")"
        else:
            node_check = f"node.has(\"{field_name_js}\") && node.get(\"{field_name_js}\")"

        if field_type.type_name == 'byte[]':
            class_definition += f"({node_check}.isBinary())"
        elif field_type.type_name == 'string' or field_type.type_name == 'String':
            class_definition += f"({node_check}.isTextual())"
        elif field_type.type_name == 'int' or field_type.type_name == 'Integer':
            class_definition += f"({node_check}.canConvertToInt())"
        elif field_type.type_name == 'long' or field_type.type_name == 'Long':
            class_definition += f"({node_check}.canConvertToLong())"
        elif field_type.type_name == 'float' or field_type.type_name == 'Float':
            class_definition += f"({node_check}.isFloat())"
        elif field_type.type_name == 'double' or field_type.type_name == 'Double':
            class_definition += f"({node_check}.isDouble())"
        elif field_type.type_name == 'BigDecimal':
            class_definition += f"({node_check}.isBigDecimal())"
        elif field_type.type_name == 'boolean' or field_type.type_name == 'Boolean':
            class_definition += f"({node_check}.isBoolean())"
        elif field_type.type_name == 'UUID':
            class_definition += f"({node_check}.isTextual())"
        elif field_type.type_name == 'LocalDate':
            class_definition += f"({node_check}.isTextual())"
        elif field_type.type_name == 'LocalTime':
            class_definition += f"({node_check}.isTextual())"
        elif field_type.type_name == 'Instant':
            class_definition += f"({node_check}.isTextual())"
        elif field_type.type_name == 'LocalDateTime':
            class_definition += f"({node_check}.isTextual())"
        elif field_type.type_name == 'Duration':
            class_definition += f"({node_check}.isTextual())"
        elif field_type.type_name == "Object":
            class_definition += f"({node_check}.isObject())"   
        elif field_type.type_name.startswith("List<"):
            items_type = field_type.type_name[5:-1]
            pred = f"Predicate<JsonNode> val{field_name_js} = (JsonNode n) -> n.isArray() && !n.elements().hasNext() || "
            pred_test = self.predicate_test(items_type)
            if pred_test:
                pred += "n.elements().next()" + pred_test
            elif items_type in self.generated_types_java_package:
                kind = self.generated_types_java_package[items_type]
                if kind == "enum":
                    pred += f"n.elements().next().isTextual() && Enum.valueOf({items_type}.class, n.elements().next().asText()) != null"
                else:
                    pred += f"{items_type}.isJsonMatch(n.elements().next())"
            else:
                pred += "true"
            predicates += pred + ";"
            class_definition += f"(node.has(\"{field_name_js}\") && val{field_name_js}.test(node.get(\"{field_name_js}\")))"
        elif field_type.type_name.startswith("Map<"):
            comma_offset = field_type.type_name.find(',')+1
            values_type = field_type.type_name[comma_offset:-1]
            pred = f"Predicate<JsonNode> val{field_name_js} = (JsonNode n) -> n.isObject() && !n.elements().hasNext() || "
            pred_test = self.predicate_test(values_type)
            if pred_test:
                pred += "n.elements().next()" + pred_test
            elif values_type in self.generated_types_java_package:
                kind = self.generated_types_java_package[values_type]
                if kind == "enum":
                    pred += f"n.elements().next().isTextual() && Enum.valueOf({values_type}.class, n.elements().next().asText()) != null"
                else:
                    pred += f"{values_type}.isJsonMatch(n.elements().next())"
            else:
                pred += "true"
            predicates += pred + ";"
            class_definition += f"(node.has(\"{field_name_js}\") && val{field_name_js}.test(node.get(\"{field_name_js}\")))"
        elif field_type.is_class:
            class_definition += f"(node.has(\"{field_name_js}\") && {field_type.type_name}.isJsonMatch(node.get(\"{field_name_js}\")))"
        elif field_type.is_enum:
            class_definition += f"(node.get(\"{field_name_js}\").isTextual() && Enum.valueOf({field_type.type_name}.class, node.get(\"{field_name_js}\").asText()) != null)"
        else:
            is_union = False
            field_union = pascal(field_name) + 'Union'
            if field_type == field_union:
                field_union = class_name + "." + pascal(field_name) + 'Union'
                type_kind = self.generated_types_avro_namespace[field_union] if field_union in self.generated_types_avro_namespace else "class"
                if type_kind == "union":
                    is_union = True
                    class_definition += f"({node_check}.isObject() && {field_type.type_name}.isJsonMatch(node.get(\"{field_name_js}\")))"
            if not is_union:
                class_definition += f"(node.has(\"{field_name_js}\"))"
        return predicates, class_definition

    def predicate_test(self, items_type):
        """ Generates the predicate test for a list or map"""
        if items_type == "String":
            return ".isTextual()"
        elif items_type in ['int', 'Integer']:
            return ".canConvertToInt()"
        elif items_type in ['long', 'Long']:
            return ".canConvertToLong()"
        elif items_type in ['float', 'Float', 'double', 'Double', 'decimal']:
            return ".isNumber()"
        elif items_type in ['boolean', 'Boolean']:
            return ".isBoolean()"
        elif items_type == 'byte[]':
            return ".isBinary()"
        elif items_type == 'UUID':
            return ".isTextual()"
        elif items_type == 'LocalDate':
            return ".isTextual()"
        elif items_type == 'LocalTime':
            return ".isTextual()"
        elif items_type == 'Instant':
            return ".isTextual()"
        elif items_type == 'LocalDateTime':
            return ".isTextual()"
        elif items_type == 'Duration':
            return ".isTextual()"
        elif items_type == "Object":
            return ".isObject()"
        return ""
    
    def get_is_json_match_clause_type(self, element_name: str, class_name: str, field_type: JavaType) -> str:
        """ Generates the isJsonMatch clause for a field using Jackson """
        predicates = ''
        class_definition = ''
        is_optional = field_type.type_name[-1] == '?'
        #is_optional = field_type[-1] == '?'
        #field_type = field_type[:-1] if is_optional else field_type
        is_optional = False
        node_check = f"{element_name}.isMissingNode() == false && {element_name}"
        null_check = f"{element_name}.isNull()" if is_optional else "false"
        if field_type.type_name == 'byte[]':
            class_definition += f"({node_check}.isBinary()){f' || {null_check}' if is_optional else ''}"
        elif field_type.type_name == 'String':
            class_definition += f"({node_check}.isTextual()){f' || {null_check}' if is_optional else ''}"
        elif self.is_java_numeric_type(field_type):
            class_definition += f"({node_check}.isNumber()){f' || {null_check}' if is_optional else ''}"
        elif field_type.type_name == 'bool' or field_type.type_name == 'Boolean':
            class_definition += f"({node_check}.isBoolean()){f' || {null_check}' if is_optional else ''}"
        elif field_type.type_name.startswith("List<"):
            items_type = field_type.type_name[5:-1]
            predicates += f"Predicate<JsonNode> val{element_name}. = (JsonNode n) -> n.isObject() && n.fields().hasNext() && n.fields().next().getValue().isTextual();"
            class_definition += f"({node_check}.isArray()){f' || {null_check}' if is_optional else ''}"
        elif field_type.type_name.startswith("Map<"):
            values_type = field_type.type_name[4:-1]
            class_definition += f"({node_check}.isObject()){f' || {null_check}' if is_optional else ''}"
        elif field_type.is_class:
            class_definition += f"({null_check} || {field_type.type_name}.isJsonMatch({element_name}))"
        elif field_type.is_enum:
            class_definition += f"({null_check} || ({node_check}.isTextual() && Enum.valueOf({field_type.type_name}.class, {element_name}.asText()) != null))"
        else:
            is_union = False
            field_union = pascal(element_name) + 'Union'
            if field_type == field_union:
                field_union = class_name + "." + pascal(element_name) + 'Union'
                type_kind = self.generated_types_avro_namespace[field_union] if field_union in self.generated_types_avro_namespace else "class"
                if type_kind == "union":
                    is_union = True
                    class_definition += f"({null_check} || {field_type}.isJsonMatch({element_name}))"
            if not is_union:
                class_definition += f"({node_check}.isObject()){f' || {null_check}' if is_optional else ''}"

        return class_definition

    def generate_avro_get_method(self, class_name: str, fields: List[Dict], parent_package: str) -> str:
        """ Generates the get method for SpecificRecord """
        get_method = f"\n{INDENT}@Override\n{INDENT}public Object get(int field$) {{\n"
        get_method += f"{INDENT * 2}switch (field$) {{\n"
        for index, field in enumerate(fields):
            field_name = pascal(field['name']) if self.pascal_properties else field['name']
            field_name = self.safe_identifier(field_name, class_name)
            field_type = self.convert_avro_type_to_java(class_name, field_name, field['type'], parent_package)
            if field_type.type_name in self.generated_types_avro_namespace and self.generated_types_avro_namespace[field_type.type_name] == "union":
                get_method += f"{INDENT * 3}case {index}: return this.{field_name}!=null?this.{field_name}.toObject():null;\n"
            else:
                get_method += f"{INDENT * 3}case {index}: return this.{field_name};\n"
        get_method += f"{INDENT * 3}default: throw new AvroRuntimeException(\"Bad index: \" + field$);\n"
        get_method += f"{INDENT * 2}}}\n{INDENT}}}\n"
        return get_method

    def generate_avro_put_method(self, class_name: str, fields: List[Dict], parent_package: str) -> str:
        """ Generates the put method for SpecificRecord """
        suppress_unchecked = False
        put_method = f"\n{INDENT}@Override\n{INDENT}public void put(int field$, Object value$) {{\n"
        put_method += f"{INDENT * 2}switch (field$) {{\n"
        for index, field in enumerate(fields):
            field_name = pascal(field['name']) if self.pascal_properties else field['name']
            field_name = self.safe_identifier(field_name, class_name)
            field_type = self.convert_avro_type_to_java(class_name, field_name, field['type'], parent_package)
            if field_type.type_name.startswith("List<") or field_type.type_name.startswith("Map<"):
                suppress_unchecked = True
            if field_type.type_name in self.generated_types_avro_namespace and self.generated_types_avro_namespace[field_type.type_name] == "union":
                put_method += f"{INDENT * 3}case {index}: this.{field_name} = new {field_type.type_name}((GenericData.Record)value$); break;\n"
            else:
                if field_type.type_name == 'String':
                    put_method += f"{INDENT * 3}case {index}: this.{field_name} = value$.toString(); break;\n"
                else:
                    put_method += f"{INDENT * 3}case {index}: this.{field_name} = ({field_type.type_name})value$; break;\n"
        put_method += f"{INDENT * 3}default: throw new AvroRuntimeException(\"Bad index: \" + field$);\n"
        put_method += f"{INDENT * 2}}}\n{INDENT}}}\n"
        if suppress_unchecked:
            put_method = f"\n{INDENT}@SuppressWarnings(\"unchecked\"){put_method}"
        return put_method

    def generate_enum(self, avro_schema: Dict, parent_package: str, write_file: bool) -> JavaType:
        """ Generates a Java enum from an Avro enum schema """
        enum_definition = ''
        if 'doc' in avro_schema:
            enum_definition += f"/** {avro_schema['doc']} */\n"
            
        package = self.join_packages(self.base_package, avro_schema.get('namespace', parent_package)).replace('.', '/').lower()       
        enum_name = self.safe_identifier(avro_schema['name'])
        type_name = self.qualified_name(package.replace('/', '.'), enum_name)
        self.generated_types_avro_namespace[self.qualified_name(avro_schema.get('namespace', parent_package),avro_schema['name'])] = "enum"
        self.generated_types_java_package[type_name] = "enum"       
        symbols = avro_schema.get('symbols', [])
        symbols_str = ', '.join([symbol.upper() for symbol in symbols])
        enum_definition += f"public enum {enum_name} {{\n"
        enum_definition += f"{INDENT}{symbols_str};\n"
        enum_definition += "}\n"
        if write_file:
            self.write_to_file(package, enum_name, enum_definition)
        return AvroToJava.JavaType(type_name, is_enum=True)
    
    def generate_embedded_union_class_jackson(self, class_name: str, field_name: str, avro_type: List, parent_package: str, write_file: bool) -> str:
        """ Generates an embedded Union Class for Java using Jackson """
        class_definition_ctors = class_definition_decls = class_definition_read = class_definition_write = class_definition = ''
        class_definition_toobject = class_definition_fromobjectctor = class_definition_genericrecordctor = ''
        
        list_is_json_match: List[str] = []
        union_class_name = class_name + pascal(field_name) + 'Union'
        package = self.join_packages(self.base_package, parent_package).replace('.', '/').lower()
        union_types: List[AvroToJava.JavaType] = [self.convert_avro_type_to_java(class_name, field_name + "Option" + str(i), t, parent_package) for i, t in enumerate(avro_type)]
        for i, union_type in enumerate(union_types):
            # we need the nullable version (wrapper) of all primitive types
            if self.is_java_primitive(union_type):
                union_type = self.map_primitive_to_java(union_type.type_name, True)
            union_variable_name = union_type.type_name
            is_dict = is_list = False
            if union_type.type_name.startswith("Map<"):
                # handle Map types
                is_dict = True
                # find the comma
                union_variable_name = flatten_type_name(union_type.type_name)
            elif union_type.type_name.startswith("List<"):
                # handle List types
                is_list = True
                union_variable_name = flatten_type_name(union_type.type_name)
            elif union_type.type_name == "byte[]":
                union_variable_name = "Bytes"
            else:
                union_variable_name = union_type.type_name.rsplit('.', 1)[-1]
            
            union_variable_name = self.safe_identifier(union_variable_name, class_name)

            # Constructor for each type
            class_definition_ctors += \
                f"{INDENT*1}public {union_class_name}({union_type.type_name} {union_variable_name}) {{\n{INDENT*2}this._{camel(union_variable_name)} = {union_variable_name};\n{INDENT*1}}}\n"

            # Declarations
            class_definition_decls += \
                f"{INDENT*1}private {union_type.type_name} _{camel(union_variable_name)};\n" + \
                f"{INDENT*1}public {union_type.type_name} get{union_variable_name}() {{ return _{camel(union_variable_name)}; }}\n";
                
            class_definition_toobject += f"{INDENT*2}if (_{camel(union_variable_name)} != null) {{\n{INDENT*3}return _{camel(union_variable_name)};\n{INDENT*2}}}\n"
            
            if self.avro_annotation and union_type.is_class:            
                class_definition_genericrecordctor += f"{INDENT*2}if ( {union_type.type_name}.AVROSCHEMA.getName().equals(record.getSchema().getName()) && {union_type.type_name}.AVROSCHEMA.getNamespace().equals(record.getSchema().getNamespace()) ) {{"
                class_definition_genericrecordctor += f"\n{INDENT*3}this._{camel(union_variable_name)} = new {union_type.type_name}(record);\n{INDENT*3}return;\n{INDENT*2}}}\n"
            
            # there can only be one list and one map in the union, so we don't need to differentiate this any further
            if is_list:
                class_definition_fromobjectctor += f"{INDENT*2}if (obj instanceof List<?>) {{\n{INDENT*3}this._{camel(union_variable_name)} = ({union_type.type_name})obj;\n{INDENT*3}return;\n{INDENT*2}}}\n"
            elif is_dict:
                class_definition_fromobjectctor += f"{INDENT*2}if (obj instanceof Map<?,?>) {{\n{INDENT*3}this._{camel(union_variable_name)} = ({union_type.type_name})obj;\n{INDENT*3}return;\n{INDENT*2}}}\n"
            else:
                class_definition_fromobjectctor += f"{INDENT*2}if (obj instanceof {union_type.type_name}) {{\n{INDENT*3}this._{camel(union_variable_name)} = ({union_type.type_name})obj;\n{INDENT*3}return;\n{INDENT*2}}}\n"

            # Read method logic
            if is_dict:
                class_definition_read += f"{INDENT*3}if (node.isObject()) {{\n{INDENT*4}{union_type.type_name} map = mapper.readValue(node.toString(), new TypeReference<{union_type.type_name}>(){{}});\n{INDENT*3}return new {union_class_name}(map);\n{INDENT*3}}}\n"
            elif is_list:
                class_definition_read += f"{INDENT*3}if (node.isArray()) {{\n{INDENT*4}{union_type.type_name} list = mapper.readValue(node.toString(), new TypeReference<{union_type.type_name}>(){{}});\n{INDENT*4}return new {union_class_name}(list);\n{INDENT*3}}}\n"
            elif self.is_java_primitive(union_type):
                if union_type.type_name == "String":
                    class_definition_read += f"{INDENT*3}if (node.isTextual()) {{\n{INDENT*4}return new {union_class_name}(node.asText());\n{INDENT*3}}}\n"
                elif union_type.type_name == "byte[]":
                    class_definition_read += f"{INDENT*3}if (node.isBinary()) {{\n{INDENT*4}return new {union_class_name}(node.binaryValue());\n{INDENT*3}}}\n"
                elif union_type.type_name in ["int", "Int"]:
                    class_definition_read += f"{INDENT*3}if (node.canConvertToInt()) {{\n{INDENT*4}return new {union_class_name}(node.asInt());\n{INDENT*3}}}\n"
                elif union_type.type_name in ["long", "Long"]:
                    class_definition_read += f"{INDENT*3}if (node.canConvertToLong()) {{\n{INDENT*4}return new {union_class_name}(node.asLong());\n{INDENT*3}}}\n"
                elif union_type.type_name in ["float", "Float"]:
                    class_definition_read += f"{INDENT*3}if (node.isFloat()) {{\n{INDENT*4}return new {union_class_name}(node.floatValue());\n{INDENT*3}}}\n"
                elif union_type.type_name in ["double", "Double"]:
                    class_definition_read += f"{INDENT*3}if (node.isDouble()) {{\n{INDENT*4}return new {union_class_name}(node.doubleValue());\n{INDENT*3}}}\n"
                elif union_type.type_name == "decimal":
                    class_definition_read += f"{INDENT*3}if (node.isBigDecimal()) {{\n{INDENT*4}return new {union_class_name}(node.decimalValue());\n{INDENT*3}}}\n"
                elif union_type.type_name in ["boolean", "Boolean"]:
                    class_definition_read += f"{INDENT*3}if (node.isBoolean()) {{\n{INDENT*4}return new {union_class_name}(node.asBoolean());\n{INDENT*3}}}\n"
            else:
                if union_type.is_enum:
                    class_definition_read += f"{INDENT*3}if (node.isTextual()) {{\n{INDENT*4}return new {union_class_name}(Enum.valueOf({union_type.type_name}.class, node.asText()));\n{INDENT*3}}}\n"
                else:
                    class_definition_read += f"{INDENT*3}if (node.isObject() && {union_type.type_name}.isJsonMatch(node)) {{\n{INDENT*4}return new {union_class_name}(mapper.readValue(node.toString(), {union_type.type_name}.class));\n{INDENT*3}}}\n"
                
            # Write method logic
            class_definition_write += f"{INDENT*3}{union_type.type_name} {camel(union_variable_name)}Value = value.get{union_variable_name}();\n{INDENT*3}if ({camel(union_variable_name)}Value != null) {{\n{INDENT*4}generator.writeObject({camel(union_variable_name)}Value);\n{INDENT*4}return;\n{INDENT*3}}}\n"

            # JSON match method logic
            gij = self.get_is_json_match_clause_type("node", class_name, union_type)
            if gij:
                list_is_json_match.append(gij)

        class_definition =  f"@JsonSerialize(using = {union_class_name}.Serializer.class)\n"
        class_definition += f"@JsonDeserialize(using = {union_class_name}.Deserializer.class)\n"
        class_definition += f"public class {union_class_name} {{\n"
        class_definition += class_definition_decls
        class_definition += f"\n{INDENT}public " + union_class_name + "() {}\n"
        if self.avro_annotation:
            class_definition += f"\n{INDENT}public {union_class_name}(GenericData.Record record) {{\n"
            class_definition += class_definition_genericrecordctor
            class_definition += f"{INDENT*2}throw new UnsupportedOperationException(\"No record type is set in the union\");\n"
            class_definition += f"{INDENT}}}\n"
        class_definition += f"\n{INDENT}public {union_class_name}(Object obj) {{\n"
        class_definition += class_definition_fromobjectctor
        class_definition += f"{INDENT*2}throw new UnsupportedOperationException(\"No record type is set in the union\");\n"
        class_definition += f"{INDENT}}}\n"
        class_definition += class_definition_ctors
        class_definition += f"\n{INDENT}public Object toObject() {{\n"
        class_definition += class_definition_toobject
        class_definition += f"{INDENT*2}throw new UnsupportedOperationException(\"No record type is set in the union\");\n"
        class_definition += f"{INDENT}}}\n"
        class_definition += f"\n{INDENT}public static class Serializer extends JsonSerializer<" + union_class_name + "> {\n"
        class_definition += f"{INDENT*2}@Override\n"
        class_definition += f"{INDENT*2}public void serialize(" + union_class_name + " value, JsonGenerator generator, SerializerProvider serializers) throws IOException {\n"
        class_definition += class_definition_write
        class_definition += f"{INDENT*3}throw new UnsupportedOperationException(\"No record type is set in the union\");\n"
        class_definition += f"{INDENT*2}}}\n{INDENT}}}\n"
        class_definition += f"\n{INDENT}public static class Deserializer extends JsonDeserializer<" + union_class_name + "> {\n"
        class_definition += f"{INDENT*2}@Override\n"
        class_definition += f"{INDENT*2}public " + union_class_name + " deserialize(JsonParser p, DeserializationContext ctxt) throws IOException, JsonProcessingException {\n"
        class_definition += f"{INDENT*3}ObjectMapper mapper = (ObjectMapper) p.getCodec();\n"
        class_definition += f"{INDENT*3}JsonNode node = mapper.readTree(p);\n"
        class_definition += class_definition_read
        class_definition += f"{INDENT*3}throw new UnsupportedOperationException(\"No record type matched the JSON data\");\n"
        class_definition += f"{INDENT*2}}}\n{INDENT}}}\n"
        class_definition += f"\n{INDENT*1}public static boolean isJsonMatch(JsonNode node) {{\n"
        class_definition += f"{INDENT*2}return " + " || ".join(list_is_json_match) + ";\n"
        class_definition += f"{INDENT*1}}}\n}}\n"

        if write_file:
            self.write_to_file(package, union_class_name, class_definition)
        self.generated_types_avro_namespace[union_class_name] = "union"  # Track union types
        self.generated_types_java_package[union_class_name] = "union"  # Track union types
        return union_class_name


    def generate_property(self, class_name: str, field: Dict, parent_package: str) -> str:
        """ Generates a Java property definition """
        field_name = pascal(field['name']) if self.pascal_properties else field['name']
        field_type = self.convert_avro_type_to_java(class_name, field_name, field['type'], parent_package)
        safe_field_name = self.safe_identifier(field_name, class_name)
        property_def = ''
        if 'doc' in field:
            property_def += f"{INDENT}/** {field['doc']} */\n"
        if self.jackson_annotations:
            property_def += f"{INDENT}@JsonProperty(\"{field['name']}\")\n"
        property_def += f"{INDENT}private {field_type.type_name} {safe_field_name};\n"
        property_def += f"{INDENT}public {field_type.type_name} get{pascal(field_name)}() {{ return {safe_field_name}; }}\n"
        property_def += f"{INDENT}public void set{pascal(field_name)}({field_type.type_name} {safe_field_name}) {{ this.{safe_field_name} = {safe_field_name}; }}\n"
        if field_type.union_types:
            for union_type in field_type.union_types:
                if union_type.type_name.startswith("List<") or union_type.type_name.startswith("Map<"):
                    property_def += f"{INDENT}@SuppressWarnings(\"unchecked\")\n"
                property_def += f"{INDENT}public {union_type.type_name} get{pascal(field_name)}As{flatten_type_name(union_type.type_name)}() {{ return ({union_type.type_name}){safe_field_name}; }}\n"
                property_def += f"{INDENT}public void set{pascal(field_name)}As{flatten_type_name(union_type.type_name)}({union_type.type_name} {safe_field_name}) {{ this.{safe_field_name} = {safe_field_name}; }}\n"
        return property_def

    def write_to_file(self, package: str, name: str, definition: str):
        """ Writes a Java class or enum to a file """
        package = package.lower()
        package = self.safe_package(package)
        directory_path = os.path.join(
            self.output_dir, package.replace('.', os.sep).replace('/', os.sep))
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)
        file_path = os.path.join(directory_path, f"{name}.java")

        with open(file_path, 'w', encoding='utf-8') as file:
            if package:
                file.write(f"package {package.replace('/', '.')};\n\n")
                if "List<" in definition:
                    file.write("import java.util.List;\n")
                if "Map<" in definition:
                    file.write("import java.util.Map;\n")
                if "Predicate<" in definition:
                    file.write("import java.util.function.Predicate;\n")
                if "BigDecimal" in definition:
                    file.write("import java.math.BigDecimal;\n")
                if "LocalDate" in definition:
                    file.write("import java.time.LocalDate;\n")
                if "LocalTime" in definition:
                    file.write("import java.time.LocalTime;\n")
                if "Instant" in definition:
                    file.write("import java.time.Instant;\n")
                if "LocalDateTime" in definition:
                    file.write("import java.time.LocalDateTime;\n")
                if "UUID" in definition:
                    file.write("import java.util.UUID;\n")
                if "Duration" in definition:
                    file.write("import java.time.Duration;\n")
                    
            if self.avro_annotation:
                if 'AvroRuntimeException' in definition:
                    file.write("import org.apache.avro.AvroRuntimeException;\n")
                if 'Schema' in definition:
                    file.write("import org.apache.avro.Schema;\n")
                if 'GenericData' in definition:
                    file.write("import org.apache.avro.generic.GenericData;\n")
                if 'DatumReader' in definition:
                    file.write("import org.apache.avro.io.DatumReader;\n")
                if 'DatumWriter' in definition:
                    file.write("import org.apache.avro.io.DatumWriter;\n")
                if 'DecoderFactory' in definition:
                    file.write("import org.apache.avro.io.DecoderFactory;\n")
                if 'EncoderFactory' in definition:
                    file.write("import org.apache.avro.io.EncoderFactory;\n")
                if 'SpecificDatumReader' in definition:
                    file.write("import org.apache.avro.specific.SpecificDatumReader;\n")
                if 'SpecificDatumWriter' in definition:
                    file.write("import org.apache.avro.specific.SpecificDatumWriter;\n")
                if 'SpecificRecord' in definition:
                    file.write("import org.apache.avro.specific.SpecificRecord;\n")
                if 'Encoder' in definition:
                    file.write("import org.apache.avro.io.Encoder;\n")
            if self.jackson_annotations:
                if 'JsonNode' in definition:
                    file.write("import com.fasterxml.jackson.databind.JsonNode;\n")
                if 'ObjectMapper' in definition:
                    file.write("import com.fasterxml.jackson.databind.ObjectMapper;\n")
                if 'JsonSerialize' in definition:
                    file.write("import com.fasterxml.jackson.databind.annotation.JsonSerialize;\n")
                if 'JsonDeserialize' in definition:
                    file.write("import com.fasterxml.jackson.databind.annotation.JsonDeserialize;\n")               
                if 'JsonSerializer' in definition:
                    file.write("import com.fasterxml.jackson.databind.JsonSerializer;\n") 
                if 'SerializerProvider' in definition:
                    file.write("import com.fasterxml.jackson.databind.SerializerProvider;\n")
                if 'JsonDeserializer' in definition:
                    file.write("import com.fasterxml.jackson.databind.JsonDeserializer;\n")
                if 'DeserializationContext' in definition:
                    file.write("import com.fasterxml.jackson.databind.DeserializationContext;\n")
                if 'JsonParser' in definition:
                    file.write("import com.fasterxml.jackson.core.JsonParser;\n")
                if 'JsonIgnore' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonIgnore;\n")
                if 'JsonProperty' in definition:
                    file.write("import com.fasterxml.jackson.annotation.JsonProperty;\n")
                if 'JsonProcessingException' in definition:
                    file.write("import com.fasterxml.jackson.core.JsonProcessingException;\n")
                if 'JsonGenerator' in definition:
                    file.write("import com.fasterxml.jackson.core.JsonGenerator;\n")
                if 'TypeReference' in definition:
                    file.write("import com.fasterxml.jackson.core.type.TypeReference;\n")
            if self.avro_annotation or self.jackson_annotations:
                if 'GZIPOutputStream' in definition:
                    file.write("import java.util.zip.GZIPOutputStream;\n")
                if 'GZIPInputStream' in definition:
                    file.write("import java.util.zip.GZIPInputStream;\n")
                if 'ByteArrayInputStream' in definition:
                    file.write("import java.io.ByteArrayInputStream;\n")
                if "ByteArrayOutputStream" in definition:
                    file.write("import java.io.ByteArrayOutputStream;\n")
                if "InputStream" in definition:
                    file.write("import java.io.InputStream;\n")
                if "IOException" in definition:
                    file.write("import java.io.IOException;\n")
                if "InflaterInputStream" in definition:
                    file.write("import java.util.zip.InflaterInputStream;\n")                
            file.write("\n")
            file.write(definition)

    def convert_schema(self, schema: JsonNode, output_dir: str):
        """Converts Avro schema to Java"""
        if not isinstance(schema, list):
            schema = [schema]
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        pom_path = os.path.join(output_dir, "pom.xml")
        if not os.path.exists(pom_path):
            package_elements = self.base_package.split('.') if self.base_package else ["com", "example"]
            groupid = '.'.join(package_elements[:-1]) if len(package_elements) > 1 else package_elements[0]
            artifactid = package_elements[-1]
            with open(pom_path, 'w', encoding='utf-8') as file:
                file.write(POM_CONTENT.format(groupid=groupid, artifactid=artifactid, AVRO_VERSION=AVRO_VERSION, JACKSON_VERSION=JACKSON_VERSION, JDK_VERSION=JDK_VERSION, PACKAGE=self.base_package))
        output_dir = os.path.join(
            output_dir, "src/main/java".replace('/', os.sep))
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        self.output_dir = output_dir
        for avro_schema in (x for x in schema if isinstance(x, dict)):
            self.generate_class_or_enum(avro_schema, '')

    def convert(self, avro_schema_path: str, output_dir: str):
        """Converts Avro schema to Java"""
        with open(avro_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        self.convert_schema(schema, output_dir)


def convert_avro_to_java(avro_schema_path, java_file_path, package_name='', pascal_properties=False, jackson_annotation=False, avro_annotation=False):
    """_summary_

    Converts Avro schema to C# classes

    Args:
        avro_schema_path (_type_): Avro input schema path  
        cs_file_path (_type_): Output C# file path 
    """
    if not package_name:
        package_name = os.path.splitext(os.path.basename(java_file_path))[0].replace('-', '_').lower()
    avrotojava = AvroToJava()
    avrotojava.base_package = package_name
    avrotojava.pascal_properties = pascal_properties
    avrotojava.avro_annotation = avro_annotation
    avrotojava.jackson_annotations = jackson_annotation
    avrotojava.convert(avro_schema_path, java_file_path)


def convert_avro_schema_to_java(avro_schema: JsonNode, output_dir: str, package_name='', pascal_properties=False, jackson_annotation=False, avro_annotation=False):
    """_summary_

    Converts Avro schema to C# classes

    Args:
        avro_schema (_type_): Avro schema as a dictionary or list of dictionaries
        output_dir (_type_): Output directory path 
    """
    avrotojava = AvroToJava()
    avrotojava.base_package = package_name
    avrotojava.pascal_properties = pascal_properties
    avrotojava.avro_annotation = avro_annotation
    avrotojava.jackson_annotations = jackson_annotation
    avrotojava.convert_schema(avro_schema, output_dir)



"""Convert an Avro schema to an Iceberg schema."""

import json
import sys
from typing import Dict, List
import pyarrow as pa
from pyiceberg.schema import Schema, NestedField
from pyiceberg.types import (
    BooleanType,
    IntegerType,
    LongType,
    FloatType,
    DoubleType,
    StringType,
    BinaryType,
    DateType,
    TimestampType,
    DecimalType,
    FixedType,
    ListType,
    MapType,
    StructType
)
from pyiceberg.io.pyarrow import PyArrowFileIO, schema_to_pyarrow

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | bool | int | None


class AvroToIcebergConverter:
    """Class to convert Avro schema to Iceberg schema."""

    def __init__(self: 'AvroToIcebergConverter'):
        self.named_type_cache: Dict[str, JsonNode] = {}
        self.id_counter = 0

    def get_id(self) -> int:
        """Get a unique ID for a record type."""
        self.id_counter += 1
        return self.id_counter
    
    def get_fullname(self, namespace: str, name: str) -> str:
        """Get the full name of a record type."""
        return f"{namespace}.{name}" if namespace else name

    def convert_avro_to_iceberg(self, avro_schema_path: str, avro_record_type: str, output_path: str, emit_cloudevents_columns: bool=False):
        """Convert an Avro schema to an Iceberg schema."""
        schema_file = avro_schema_path
        if not schema_file:
            print("Please specify the avro schema file")
            sys.exit(1)
        with open(schema_file, "r", encoding="utf-8") as f:
            schema_json = f.read()

        # Parse the schema as a JSON object
        schema = json.loads(schema_json)
        self.cache_named_types(schema)

        if isinstance(schema, list) and avro_record_type:
            schema = next(
                (x for x in schema if x["name"] == avro_record_type or x["namespace"] + "." + x["name"] == avro_record_type), None)
            if schema is None:
                print(
                    f"No top-level record type {avro_record_type} found in the Avro schema")
                sys.exit(1)
        elif not isinstance(schema, dict):
            print(
                "Expected a single Avro schema as a JSON object, or a list of schema records")
            sys.exit(1)

        # Get the name and fields of the top-level record
        table_name = schema["name"]
        fields = schema["fields"]

        # Create a list to store the iceberg schema
        iceberg_fields: List[NestedField] = []

        # Append the iceberg schema with the column names and types
        for i, field in enumerate(fields):
            column_name = field["name"]
            column_type = self.convert_avro_type_to_iceberg_type(field["type"])
            iceberg_fields.append(
                NestedField(field_id=self.get_id(), name=column_name, type=column_type))

        if emit_cloudevents_columns:
            iceberg_fields.extend([
                NestedField(field_id=self.get_id(),
                      name="___type", type=StringType()),
                NestedField(field_id=self.get_id(),
                      name="___source", type=StringType()),
                NestedField(field_id=self.get_id(),
                      name="___id", type=StringType()),
                NestedField(field_id=self.get_id(),
                      name="___time", type=TimestampType()),
                NestedField(field_id=self.get_id(),
                      name="___subject", type=StringType())
            ])

        iceberg_schema = Schema(*iceberg_fields)
        arrow_schema = schema_to_pyarrow(iceberg_schema)
        print(f"Iceberg schema created: {arrow_schema}")

        # Write to Iceberg table (for demonstration, using local file system)
        file_io = PyArrowFileIO()
        output_file = file_io.new_output("file://"+output_path)
        with output_file.create(overwrite=True) as f:
            pa.output_stream(f).write(arrow_schema.serialize().to_pybytes())

    def convert_avro_type_to_iceberg_type(self, avro_type):
        """Convert an Avro type to an Iceberg type."""
        if isinstance(avro_type, list):
            item_count = len(avro_type)
            if item_count == 1:
                return self.convert_avro_type_to_iceberg_type(avro_type[0])
            elif item_count == 2:
                first, second = avro_type[0], avro_type[1]
                if first == "null":
                    return self.convert_avro_type_to_iceberg_type(second)
                elif second == "null":
                    return self.convert_avro_type_to_iceberg_type(first)
                else:
                    return StructType(fields=[NestedField(field_id=self.get_id(), name=f'field_{i}', type=self.convert_avro_type_to_iceberg_type(t)) for i, t in enumerate(avro_type)])
            elif item_count > 0:
                return StructType(fields=[NestedField(field_id=self.get_id(), name=f'field_{i}', type=self.convert_avro_type_to_iceberg_type(t)) for i, t in enumerate(avro_type)])
            else:
                print(f"WARNING: Empty union type {avro_type}")
                return StringType()
        elif isinstance(avro_type, dict):
            type_name = avro_type.get("type")
            if type_name == "array":
                return ListType(element_id=self.get_id(), element=self.convert_avro_type_to_iceberg_type(avro_type.get("items")))
            elif type_name == "map":
                return MapType(key_id=self.get_id(), key_type=StringType(), value_id=self.get_id(), value_type=self.convert_avro_type_to_iceberg_type(avro_type.get("values")))
            elif type_name == "record":
                fields = avro_type.get("fields")
                return StructType(fields=[NestedField(field_id=self.get_id(), name=field["name"], type=self.convert_avro_type_to_iceberg_type(field["type"])) for i, field in enumerate(fields)])
            if type_name == "enum":
                return StringType()
            elif type_name == "fixed":
                return FixedType(avro_type.get("size"))
            elif type_name == "string":
                logical_type = avro_type.get("logicalType")
                if logical_type == "uuid":
                    return StringType()
                return StringType()
            elif type_name == "bytes":
                logical_type = avro_type.get("logicalType")
                if logical_type == "decimal":
                    return DecimalType(38, 18)
                return BinaryType()
            elif type_name == "long":
                logical_type = avro_type.get("logicalType")
                if logical_type in ["timestamp-millis", "timestamp-micros"]:
                    return TimestampType()
                if logical_type in ["time-millis", "time-micros"]:
                    return LongType()
                return LongType()
            elif type_name == "int":
                logical_type = avro_type.get("logicalType")
                if logical_type == "date":
                    return DateType()
                return IntegerType()
            else:
                return self.map_iceberg_scalar_type(type_name)
        elif isinstance(avro_type, str):
            if avro_type in self.named_type_cache:
                return self.convert_avro_type_to_iceberg_type(self.named_type_cache[avro_type])
            return self.map_iceberg_scalar_type(avro_type)

        return StringType()

    def cache_named_types(self, avro_type):
        """Add an encountered type to the list of types."""
        if isinstance(avro_type, list):
            for item in avro_type:
                self.cache_named_types(item)
        if isinstance(avro_type, dict) and avro_type.get("name"):
            self.named_type_cache[self.get_fullname(avro_type.get(
                "namespace"), avro_type.get("name"))] = avro_type
            if "fields" in avro_type:
                for field in avro_type.get("fields"):
                    if "type" in field:
                        self.cache_named_types(field.get("type"))

    def map_iceberg_scalar_type(self, type_name: str):
        """Map an Avro scalar type to an Iceberg scalar type."""
        if type_name == "null":
            return StringType()
        elif type_name == "int":
            return IntegerType()
        elif type_name == "long":
            return LongType()
        elif type_name == "float":
            return FloatType()
        elif type_name == "double":
            return DoubleType()
        elif type_name == "boolean":
            return BooleanType()
        elif type_name == "bytes":
            return BinaryType()
        elif type_name == "string":
            return StringType()
        else:
            return StringType()


def convert_avro_to_iceberg(avro_schema_path, avro_record_type, output_path, emit_cloudevents_columns=False):
    """Convert an Avro schema to an Iceberg schema."""
    converter = AvroToIcebergConverter()
    converter.convert_avro_to_iceberg(
        avro_schema_path, avro_record_type, output_path, emit_cloudevents_columns)



# coding: utf-8
"""
Module to convert Avro schema to a GraphQL schema.
"""

import json
import os
from typing import Dict, List

from avrotize.common import get_longest_namespace_prefix

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | bool | int | None

class AvroToGraphQLConverter:
    """
    Class to convert Avro schema to GraphQL schema.
    """

    def __init__(self, avro_schema_path, graphql_schema_path):
        """
        Initialize the converter with file paths.

        :param avro_schema_path: Path to the Avro schema file.
        :param graphql_schema_path: Path to save the GraphQL schema file.
        """
        self.avro_schema_path = avro_schema_path
        self.graphql_schema_path = graphql_schema_path
        self.records = {}
        self.enums = {}
        self.fixeds = {}
        self.longest_namespace_prefix = ""

    def convert(self: 'AvroToGraphQLConverter'):
        """
        Convert Avro schema to GraphQL schema and save to file.
        """
        with open(self.avro_schema_path, 'r', encoding='utf-8') as file:
            avro_schemas: JsonNode = json.load(file)
            
        self.longest_namespace_prefix = get_longest_namespace_prefix(avro_schemas)

        if isinstance(avro_schemas, dict):
            self.extract_named_types(avro_schemas)
        elif isinstance(avro_schemas, list):
            for schema in avro_schemas:
                if isinstance(schema, dict):
                    self.extract_named_types(schema)
        else:
            raise ValueError("Expected a single Avro schema as a JSON object, or a list of schema records")

        graphql_content = self.generate_graphql()

        with open(self.graphql_schema_path, "w", encoding="utf-8") as file:
            file.write(graphql_content)
            
    def qualified_name(self, schema: Dict[str, JsonNode]) -> str:
        """
        Get the full name of a record type.
        """
        name = str(schema['name'])
        namespace = str(schema.get('namespace', ''))
        if namespace.startswith(self.longest_namespace_prefix):
            namespace = namespace[len(self.longest_namespace_prefix):].strip('.')
        return f"{namespace}_{name}" if namespace else name
    
    def extract_named_types(self, schema: Dict[str, JsonNode]):
        """
        Extract all named types (records, enums, fixed) from the schema.
        """
        if isinstance(schema, dict):
            if schema['type'] == 'record':
                self.records[self.qualified_name(schema)] = schema
            elif schema['type'] == 'enum':
                self.enums[self.qualified_name(schema)] = schema
            elif schema['type'] == 'fixed':
                self.fixeds[self.qualified_name(schema)] = schema
            if 'fields' in schema and isinstance(schema['fields'], list):
                for field in schema['fields']:
                    if isinstance(field, dict) and 'type' in field and isinstance(field['type'], dict):
                        self.extract_named_types(field['type'])
            if 'items' in schema and isinstance(schema['items'], dict):
                self.extract_named_types(schema['items'])
            if 'values' in schema and isinstance(schema['values'], dict):
                self.extract_named_types(schema['values'])
        elif isinstance(schema, list):
            for sub_schema in schema:
                self.extract_named_types(sub_schema)

    def generate_graphql(self):
        """
        Generate GraphQL content from the extracted types.

        :return: GraphQL content as a string.
        """
        graphql = []

        for record in self.records.values():
            graphql.append(self.generate_graphql_record(record))

        for enum in self.enums.values():
            graphql.append(self.generate_graphql_enum(enum))

        for fixed in self.fixeds.values():
            graphql.append(self.generate_graphql_fixed(fixed))

        return "\n".join(graphql)

    def generate_graphql_record(self, record):
        """
        Generate GraphQL content for a record.

        :param record: Record schema as a dictionary.
        :return: GraphQL content as a string.
        """
        fields = []
        for field in record['fields']:
            field_type = self.get_graphql_type(field['type'])
            if 'null' not in field['type'] and field.get('default') is not None:
                field_type = f"{field_type}!"
            fields.append(f"  {field['name']}: {field_type}")
        return f"type {record['name']} {{\n" + "\n".join(fields) + "\n}"

    def generate_graphql_enum(self, enum):
        """
        Generate GraphQL content for an enum.

        :param enum: Enum schema as a dictionary.
        :return: GraphQL content as a string.
        """
        symbols = [f"  {symbol}" for symbol in enum['symbols']]
        return f"enum {enum['name']} {{\n" + "\n".join(symbols) + "\n}"

    def generate_graphql_fixed(self, fixed):
        """
        Generate GraphQL content for a fixed type.

        :param fixed: Fixed schema as a dictionary.
        :return: GraphQL content as a string.
        """
        return f"scalar {fixed['name']}"

    def get_graphql_type(self, avro_type):
        """
        Get GraphQL type as a string.

        :param avro_type: Avro type as a string or dictionary.
        :return: GraphQL type as a string.
        """
        if isinstance(avro_type, list):
            non_null_type = next(t for t in avro_type if t != "null")
            return self.get_graphql_type(non_null_type)
        if isinstance(avro_type, dict):
            if avro_type['type'] == 'array':
                return f"[{self.get_graphql_type(avro_type['items'])}]"
            if avro_type['type'] == 'map':
                return "JSON"
            if avro_type['type'] == "record" and self.qualified_name(avro_type) in self.records:
                return self.qualified_name(avro_type)
            if avro_type['type'] == "enum" and self.qualified_name(avro_type) in self.enums:
                return self.qualified_name(avro_type)
            if avro_type['type'] == "fixed" and self.qualified_name(avro_type) in self.fixeds:
                return self.qualified_name(avro_type)
            return self.get_graphql_primitive_type(avro_type['type'])
        return self.get_graphql_primitive_type(avro_type)

    def get_graphql_primitive_type(self, avro_type):
        """
        Map Avro primitive types to GraphQL types.

        :param avro_type: Avro type as a string.
        :return: GraphQL type as a string.
        """
        
        type_mapping = {
            "string": "String",
            "bytes": "String",
            "int": "Int",
            "long": "Int",
            "float": "Float",
            "double": "Float",
            "boolean": "Boolean",
            "null": "String",
            "date": "Date",
            "timestamp-millis": "DateTime"
        }
        return type_mapping.get(avro_type, 'String')
        

def convert_avro_to_graphql(avro_schema_path, graphql_schema_path):
    """
    Convert an Avro schema file to a GraphQL schema file.

    :param avro_schema_path: Path to the Avro schema file.
    :param graphql_schema_path: Path to save the GraphQL schema file.
    """
    converter = AvroToGraphQLConverter(avro_schema_path, graphql_schema_path)
    converter.convert()



# coding: utf-8
"""
Module to convert Avro schema to a comprehensive README.md.
"""

import json
import os
from jinja2 import Environment, FileSystemLoader

from avrotize.common import render_template

class AvroToMarkdownConverter:
    """
    Class to convert Avro schema to a comprehensive README.md.
    """

    def __init__(self, avro_schema_path, markdown_path):
        """
        Initialize the converter with file paths.

        :param avro_schema_path: Path to the Avro schema file.
        :param markdown_path: Path to save the README.md file.
        """
        self.avro_schema_path = avro_schema_path
        self.markdown_path = markdown_path
        self.records = {}
        self.enums = {}
        self.fixeds = {}

    def convert(self):
        """
        Convert Avro schema to Markdown and save to file.
        """
        with open(self.avro_schema_path, 'r', encoding='utf-8') as file:
            avro_schemas = json.load(file)

        schema_name = os.path.splitext(os.path.basename(self.avro_schema_path))[0]
        if isinstance(avro_schemas, dict):
            self.extract_named_types(avro_schemas)
        elif isinstance(avro_schemas, list):
            for schema in avro_schemas:
                self.extract_named_types(schema)

        self.generate_markdown(schema_name)

    def extract_named_types(self, schema, parent_namespace: str = ''):
        """
        Extract all named types (records, enums, fixed) from the schema.
        """
        
        if isinstance(schema, dict):
            ns = schema.get('namespace', parent_namespace)
            if schema['type'] == 'record':
                self.records.setdefault(ns, []).append(schema)
            elif schema['type'] == 'enum':
                self.enums.setdefault(ns, []).append(schema)
            elif schema['type'] == 'fixed':
                self.fixeds.setdefault(ns, []).append(schema)
            if 'fields' in schema:
                for field in schema['fields']:
                    self.extract_named_types(field['type'], ns)
            if 'items' in schema:
                self.extract_named_types(schema['items'], ns)
            if 'values' in schema:
                self.extract_named_types(schema['values'], ns)
        elif isinstance(schema, list):
            for sub_schema in schema:
                self.extract_named_types(sub_schema, '')
                
    def generate_markdown(self, schema_name: str):
        """
        Generate markdown content from the extracted types using Jinja2 template.

        :param schema_name: The name of the schema file.
        :return: Markdown content as a string.
        """
        render_template("avrotomd/README.md.jinja", self.markdown_path,
            schema_name = schema_name,
            records  = self.records,
            enums = self.enums,
            fixeds = self.fixeds,
            generate_field_markdown = self.generate_field_markdown
        )

    def generate_field_markdown(self, field, level):
        """
        Generate markdown content for a single field.

        :param field: Avro field as a dictionary.
        :param level: The current level of nesting.
        :return: Markdown content as a string.
        """
        field_md = []
        heading = "#" * level
        field_md.append(f"{heading} {field['name']}\n")
        field_md.append(f"- **Type:** {self.get_avro_type(field['type'])}")
        if 'doc' in field:
            field_md.append(f"- **Description:** {field['doc']}")
        if 'default' in field:
            field_md.append(f"- **Default:** {field['default']}")
        if isinstance(field['type'], dict) and field['type'].get('logicalType'):
            field_md.append(f"- **Logical Type:** {field['type']['logicalType']}")
        if 'symbols' in field.get('type', {}):
            field_md.append(f"- **Symbols:** {', '.join(field['type']['symbols'])}")
        field_md.append("\n")
        return "\n".join(field_md)

    def get_avro_type(self, avro_type):
        """
        Get Avro type as a string.

        :param avro_type: Avro type as a string or dictionary.
        :return: Avro type as a string.
        """
        if isinstance(avro_type, list):
            return " | ".join([self.get_avro_type(t) for t in avro_type])
        if isinstance(avro_type, dict):
            type_name = avro_type.get('type', 'unknown')
            namespace = avro_type.get('namespace', '')
            if type_name in [r['name'] for r in self.records.get(namespace, [])]:
                return f"[{type_name}](#record-{namespace.lower()}-{type_name.lower()})"
            if type_name in [e['name'] for e in self.enums.get(namespace, [])]:
                return f"[{type_name}](#enum-{namespace.lower()}-{type_name.lower()})"
            if type_name in [f['name'] for f in self.fixeds.get(namespace, [])]:
                return f"[{type_name}](#fixed-{namespace.lower()}-{type_name.lower()})"
            return type_name
        return avro_type

def convert_avro_to_markdown(avro_schema_path, markdown_path):
    """
    Convert an Avro schema file to a README.md file.

    :param avro_schema_path: Path to the Avro schema file.
    :param markdown_path: Path to save the README.md file.
    """
    converter = AvroToMarkdownConverter(avro_schema_path, markdown_path)
    converter.convert()



import json
from typing import List
import asn1tools
from asn1tools.codecs.ber import Sequence, SequenceOf, Integer, Boolean, Enumerated, OctetString, IA5String, UTF8String, Date, Real, Choice, Null, SetOf, Recursive, ExplicitTag
from asn1tools.codecs.ber import BitString, VisibleString, BMPString, GeneralString, GraphicString, NumericString, PrintableString, TeletexString, UniversalString, ObjectIdentifier, Set

from avrotize.common import avro_name
from avrotize.dependency_resolver import sort_messages_by_dependencies

def asn1_type_to_avro_type(asn1_type: dict, avro_schema: list, namespace: str, parent_type_name: str | None, parent_member_name: str | None, dependencies: list)-> str | dict | None:
    """Convert an ASN.1 type to an Avro type."""
    avro_type: str | dict | list | None = None
    deps: List[str] = []

    if isinstance(asn1_type,Integer):
        avro_type = 'int'
    elif isinstance(asn1_type, Boolean):
        avro_type = 'boolean'
    elif isinstance(asn1_type, Enumerated):
        symbols = [avro_name(member) for member in asn1_type.data_to_value.keys()]
        avro_type = { 
            'type': 'enum', 
            'name': avro_name(parent_member_name if parent_member_name else asn1_type.name),
            'namespace': namespace, 
            'symbols': symbols
            }
    elif isinstance(asn1_type, Sequence) or isinstance(asn1_type, Choice) or isinstance(asn1_type, Set):
        if avro_schema and next((s for s in avro_schema if s.get('name') == asn1_type.type_name), []):
            return namespace + '.' + asn1_type.type_name
        else:
            record_name = asn1_type.type_name if asn1_type.type_name else asn1_type.name
            if record_name == 'CHOICE' or record_name == 'SEQUENCE' or record_name == 'SET':
                if parent_member_name and parent_type_name:
                    record_name = parent_type_name + parent_member_name         
                elif parent_type_name:
                    record_name = parent_type_name 
                else:
                    raise ValueError(f"Can't name record without a type name and member name")      

            fields = []
            for member in asn1_type.members if isinstance(asn1_type, Choice) else asn1_type.root_members:
                field_type = asn1_type_to_avro_type(member, avro_schema, namespace, record_name, member.name, deps)
                if isinstance(field_type, dict) and (field_type.get('type') == 'record' or field_type.get('type') == 'enum'):
                    existing_type = next((t for t in avro_schema if (isinstance(t,dict) and t.get('name') == field_type['name'] and t.get('namespace') == field_type.get('namespace')) ), None)
                    if not existing_type:
                        field_type['dependencies'] = [dep for dep in deps if dep != field_type['namespace']+'.'+field_type['name']]
                        avro_schema.append(field_type)
                    field_type = namespace + '.' + field_type.get('name','')
                    dependencies.append(field_type)
                if isinstance(asn1_type, Choice):
                    field_type = [field_type, 'null']
                fields.append({
                    'name': avro_name(member.name),
                    'type': field_type
                })
            
            avro_type = {
                'type': 'record', 
                'name': avro_name(record_name),
                'namespace': namespace,
                'fields': fields
                }
    elif isinstance(asn1_type, BitString):
        avro_type = {
            'type': 'bytes',
            'logicalType': 'bitstring'
        }
    elif isinstance(asn1_type, ObjectIdentifier):
        avro_type = 'string'
    elif isinstance(asn1_type, SequenceOf):
        record_name = asn1_type.name if asn1_type.name else parent_member_name
        if record_name == 'SEQUENCE OF':
            record_name = parent_member_name if parent_member_name else ''
        if parent_type_name:
            record_name = parent_type_name + record_name
        item_type = asn1_type_to_avro_type(asn1_type.element_type, avro_schema, namespace, record_name, 'Item', deps)
        if isinstance(item_type, dict) and not item_type.get('name'):
            item_type['name'] = asn1_type.name + 'Item'
        avro_type = {
            'type': 'array',
            'namespace': namespace, 
            'name': avro_name(record_name), 
            'items': item_type
            }
    elif isinstance(asn1_type, SetOf):
        record_name = asn1_type.name if asn1_type.name else parent_member_name
        if record_name == 'SET OF':
            record_name = parent_member_name if parent_member_name else ''                
        if parent_type_name:
            record_name = parent_type_name + record_name
        item_type = asn1_type_to_avro_type(asn1_type.element_type, avro_schema, namespace, record_name, 'Item', deps)
        if isinstance(item_type, dict) and not item_type.get('name'):
            item_type['name'] = asn1_type.name + 'Item'               
        avro_type = {
            'type': 'array',
            'namespace': namespace, 
            'name': avro_name(record_name),
            'items': item_type
            }
    elif isinstance(asn1_type, OctetString):
        avro_type = 'bytes'
    elif isinstance(asn1_type, IA5String) or isinstance(asn1_type, UTF8String) or isinstance(asn1_type, VisibleString) or isinstance(asn1_type, BMPString) or isinstance(asn1_type, GeneralString) or isinstance(asn1_type, GraphicString) or isinstance(asn1_type, NumericString) or isinstance(asn1_type, PrintableString) or isinstance(asn1_type, TeletexString) or isinstance(asn1_type, UniversalString):
        avro_type = 'string'
    elif isinstance(asn1_type, Date):
        avro_type = {'type': 'int', 'logicalType': 'date'}
    elif isinstance(asn1_type, Real):
        avro_type = 'double'
    elif isinstance(asn1_type, Null):
        avro_type = 'null'
    elif isinstance(asn1_type, Recursive):
        avro_type = asn1_type.type_name
    elif isinstance(asn1_type, ExplicitTag):
        avro_type = asn1_type_to_avro_type(asn1_type.inner, avro_schema, namespace, parent_type_name, parent_member_name, dependencies)
        if isinstance(avro_type, dict):
            avro_type['name'] = asn1_type.name if asn1_type.name else asn1_type.type_name
    else:
        raise ValueError(f"Don't know how to translate ASN.1 type '{type(asn1_type)}' to Avro")

    if len(avro_schema) > 0 and isinstance(avro_type, dict) and 'name' in avro_type:
        existing_type = next((t for t in avro_schema if (isinstance(t,dict) and t.get('name') == avro_type['name'] and t.get('namespace') == avro_type.get('namespace')) ), None)
        if existing_type:
            qualified_name = namespace + '.' + existing_type.get('name','')
            dependencies.append(qualified_name)
            return qualified_name
            
    return avro_type

def convert_asn1_to_avro_schema(asn1_spec_list: List[str]):
    """Convert ASN.1 specification to Avro schema."""
    
    try:
        spec = asn1tools.compile_files(asn1_spec_list)
    except asn1tools.ParseError as e:
        print(f"Error parsing ASN.1 files: {e.args[0]}")
        for file in asn1_spec_list:
            print(f"  {file}")
        return None
    
    avro_schema: List[dict] = []
    for module_name, module in spec.modules.items():
        for type_name, asn1_type in module.items():
            dependencies: List [str] = []
            avro_type = asn1_type_to_avro_type(asn1_type.type, avro_schema, avro_name(module_name), type_name, None, dependencies)
            if avro_type and not isinstance(avro_type, str):
                avro_type['dependencies'] = [dep for dep in dependencies if dep != avro_type['namespace'] + '.' + avro_type['name']]
                avro_schema.append(avro_type)
    
    avro_schema = sort_messages_by_dependencies(avro_schema)

    if len(avro_schema) == 1:
        return avro_schema[0]
    return avro_schema

def convert_asn1_to_avro(asn1_spec_list: List[str], avro_file_path: str):
    """Convert ASN.1 specification to Avro schema and save it to a file."""
    avro_schema = convert_asn1_to_avro_schema(asn1_spec_list)
    if avro_schema is None:
        return
    with open(avro_file_path, 'w') as file:
        json.dump(avro_schema, file, indent=4)
